@@ -65,12 +65,15 @@ enum {
   EVENT_VALCHANGE,
   EVENT_SELCHANGE,
   EVENT_UNFOCUS,
-  EVENT_CALLBACK
+  EVENT_CALLBACK,
+  EVENT_DROP
 };
 
+extern wstring dragndrop;  // drop drag-and-drop contents here
+
 typedef struct control control;
 
-typedef void (*handler_fn)(control *, int event);
+typedef void (* handler_fn)(control *, int event);
 
 struct control {
   int type;
@@ -87,7 +90,7 @@ struct control {
   * 
   * For CTRL_COLUMNS, this field MUST be null.
   */
-  char *label;
+  char * label;
  /*
   * Indicate which column(s) this control occupies. This can
   * be unpacked into starting column and column span by the
@@ -109,11 +112,18 @@ struct control {
   * state.
   */
   handler_fn handler;
+ /*
+  * Identify a drag-and-drop target control by its widget ("Window") 
+  * as due to the obscure wisdom of Windows design, drag-and-drop events 
+  * are handled completely different from other events and particularly 
+  * do not provide a "Control identifier".
+  */
+  void * widget;
  /*
   * Almost all of the above functions will find it useful to
   * be able to store a piece of `void *' data.
   */
-  void *context;
+  void * context;
   union {
     struct {
      /*
@@ -163,12 +173,12 @@ struct control {
       * pointers, each of which points to a dynamically
       * allocated string.
       */
-      string *labels;     /* `nbuttons' button labels */
+      string * labels;     /* `nbuttons' button labels */
      /*
       * This points to a dynamically allocated array,
       * with the value corresponding to each button.
       */
-      int *vals;         /* `nbuttons' entries; may be null */
+      int * vals;       /* `nbuttons' entries; may be null */
     } radio;
     struct {
      /*
@@ -188,7 +198,7 @@ struct control {
       * Height of the list box, in approximate number of lines.
       * If this is zero, the list is a drop-down list.
       */
-      int height; /* height in lines */
+      int height;       /* height in lines */
      /*
       * Percentage of the dialog-box width used by the list box.
       * If this is set to 100, the label is on its own line;
@@ -207,13 +217,13 @@ struct control {
       * `percentages' must be null.
       */
       int ncols;  /* number of columns */
-      int *percentages;   /* % width of each column */
+      int * percentages;   /* % width of each column */
     } listbox;
 
     struct {
      /* In this variant, `label' MUST be null. */
-      int ncols;  /* number of columns */
-      int *percentages;   /* % width of each column */
+      int ncols;                /* number of columns */
+      int * percentages;        /* % width of each column */
      /*
       * Every time this control type appears, exactly one of
       * `ncols' and the previous number of columns MUST be one.
@@ -229,7 +239,7 @@ struct control {
   };
 
   /* Space for storing platform-specific control data */
-  void *plat_ctrl;
+  void * plat_ctrl;
 };
 
 #undef STANDARD_PREFIX
@@ -241,12 +251,12 @@ struct control {
  * in the config will be a container box within a panel.
  */
 typedef struct {
-  char *pathname;       /* panel path, e.g. "SSH/Tunnels" */
-  char *boxtitle;       /* title of container box */
-  int ncolumns; /* current no. of columns at bottom */
+  char * pathname;      /* panel path, e.g. "SSH/Tunnels" */
+  char * boxtitle;      /* title of container box */
+  int ncolumns;         /* current no. of columns at bottom */
   int ncontrols;        /* number of `control' in array */
-  int ctrlsize; /* allocated size of array */
-  control **ctrls;      /* actual array */
+  int ctrlsize;         /* allocated size of array */
+  control * * ctrls;    /* actual array */
 } controlset;
 
 /*
@@ -254,26 +264,27 @@ typedef struct {
  * controls.
  */
 typedef struct {
-  int nctrlsets;        /* number of ctrlsets */
-  int ctrlsetsize;      /* ctrlset size */
-  controlset **ctrlsets;        /* actual array of ctrlsets */
+  int nctrlsets;                /* number of ctrlsets */
+  int ctrlsetsize;              /* ctrlset size */
+  controlset * * ctrlsets;      /* actual array of ctrlsets */
   int nfrees;
   int freesize;
-  void **frees; /* array of aux data areas to free */
+  void * * frees;               /* array of aux data areas to free */
 } controlbox;
 
-controlbox *ctrl_new_box(void);
-void ctrl_free_box(controlbox *);
+extern controlbox * ctrl_new_box(void);
+extern void ctrl_free_box(controlbox *);
 
 /*
  * Standard functions used for populating a controlbox structure.
  */
 
 /* Create a controlset. */
-controlset *ctrl_new_set(controlbox *, char *path, char *title);
-void ctrl_free_set(controlset *);
+extern controlset * ctrl_new_set(controlbox *, char * path, 
+                                 char * panel, char * title);
+extern void ctrl_free_set(controlset *);
 
-void ctrl_free(control *);
+extern void ctrl_free(control *);
 
 /*
  * This function works like `malloc', but the memory it returns
@@ -283,7 +294,7 @@ void ctrl_free(control *);
  * to hold modifiable per-instance things. It's mostly here for
  * allocating structures to be passed as control handler params.
  */
-void *ctrl_alloc(controlbox *, size_t size);
+extern void * ctrl_alloc(controlbox *, size_t size);
 
 /*
  * Individual routines to create `control' structures in a controlset.
@@ -295,67 +306,70 @@ void *ctrl_alloc(controlbox *, size_t size);
  */
 
 /* `ncolumns' is followed by that many percentages, as integers. */
-control *ctrl_columns(controlset *, int ncolumns, ...);
-control *ctrl_editbox(controlset *, char *label, int percentage,
-                      handler_fn handler, void *context);
-control *ctrl_combobox(controlset *, char *label, int percentage,
-                       handler_fn handler, void *context);
-control *ctrl_listbox(controlset *, char *label, int lines, int percentage,
-                       handler_fn handler, void *context);
+extern control * ctrl_columns(controlset *, int ncolumns, ...);
+extern control * ctrl_editbox(controlset *, char * label, int percentage,
+                              handler_fn handler, void * context);
+extern control * ctrl_combobox(controlset *, char * label, int percentage,
+                               handler_fn handler, void * context);
+extern control * ctrl_listbox(controlset *, char * label, int lines, int percentage,
+                              handler_fn handler, void * context);
 /*
  * `ncolumns' is followed by (alternately) radio button titles and
  * integers, until a null in place of a title string is seen.
  */
-control *ctrl_radiobuttons(controlset *, char *label, int ncolumns,
-                           handler_fn handler, void *context, ...);
-control *ctrl_pushbutton(controlset *, char *label,
-                         handler_fn handler, void *context);
-control *ctrl_droplist(controlset *, char *label, int percentage,
-                       handler_fn handler, void *context);
-control *ctrl_fontsel(controlset *, char *label,
-                      handler_fn handler, void *context);
-control *ctrl_checkbox(controlset *, char *label,
-                       handler_fn handler, void *context);
+extern control * ctrl_radiobuttons(controlset *, char * label, int ncolumns,
+                                   handler_fn handler, void * context, ...);
+extern control * ctrl_pushbutton(controlset *, char * label,
+                                 handler_fn handler, void * context);
+extern control * ctrl_droplist(controlset *, char * label, int percentage,
+                               handler_fn handler, void * context);
+extern control * ctrl_fontsel(controlset *, char * label,
+                              handler_fn handler, void * context);
+extern control * ctrl_checkbox(controlset *, char * label,
+                               handler_fn handler, void * context);
 
 /*
  * Standard handler routines to cover most of the common cases in
  * the config box.
  */
 
-void dlg_stdcheckbox_handler(control *, int event);
-void dlg_stdstringbox_handler(control *, int event);
-void dlg_stdintbox_handler(control *, int event);
-void dlg_stdradiobutton_handler(control *, int event);
-void dlg_stdfontsel_handler(control *, int event);
-void dlg_stdcolour_handler(control *, int event);
+extern void dlg_stdcheckbox_handler(control *, int event);
+extern void dlg_stdstringbox_handler(control *, int event);
+extern void dlg_stdintbox_handler(control *, int event);
+extern void dlg_stdradiobutton_handler(control *, int event);
+extern void dlg_stdfontsel_handler(control *, int event);
+extern void dlg_stdcolour_handler(control *, int event);
 
 /*
  * Routines the platform-independent dialog code can call to read
  * and write the values of controls.
  */
-void dlg_radiobutton_set(control *, int whichbutton);
-int dlg_radiobutton_get(control *);
-void dlg_checkbox_set(control *, bool);
-bool dlg_checkbox_get(control *);
-void dlg_editbox_set(control *, string);
-void dlg_editbox_set_w(control *, wstring);
-void dlg_editbox_get(control *, string *);
-void dlg_editbox_get_w(control *, wstring *);
+extern void dlg_radiobutton_set(control *, int whichbutton);
+extern int dlg_radiobutton_get(control *);
+extern void dlg_checkbox_set(control *, bool);
+extern bool dlg_checkbox_get(control *);
+extern void dlg_editbox_set(control *, string);
+extern void dlg_editbox_set_w(control *, wstring);
+extern void dlg_editbox_get(control *, string *);
+extern void dlg_editbox_get_w(control *, wstring *);
 /* The `listbox' functions also apply to combo boxes. */
-void dlg_listbox_clear(control *);
-void dlg_listbox_add(control *, string);
-void dlg_listbox_add_w(control *, wstring);
-void dlg_fontsel_set(control *, font_spec *);
-void dlg_fontsel_get(control *, font_spec *);
+extern void dlg_listbox_clear(control *);
+extern void dlg_listbox_add(control *, string);
+extern void dlg_listbox_add_w(control *, wstring);
+extern int dlg_listbox_getcur(control *);
+extern void dlg_fontsel_set(control *, font_spec *);
+extern void dlg_fontsel_get(control *, font_spec *);
+/* Special for font sample */
+extern void dlg_text_paint(control *ctrl);
 /*
  * Set input focus into a particular control.
  */
-void dlg_set_focus(control *);
+extern void dlg_set_focus(control *);
 /*
  * This function signals to the front end that the dialog's
  * processing is completed.
  */
-void dlg_end(void);
+extern void dlg_end(void);
 
 /*
  * Routines to manage a (per-platform) colour selector.
@@ -370,8 +384,8 @@ void dlg_end(void);
  * dlg_coloursel_start() accepts an RGB triple which is used to
  * initialise the colour selector to its starting value.
  */
-void dlg_coloursel_start(colour);
-int dlg_coloursel_results(colour *);
+extern void dlg_coloursel_start(colour);
+extern int dlg_coloursel_results(colour *);
 
 /*
  * This routine is used by the platform-independent code to
@@ -382,7 +396,7 @@ int dlg_coloursel_results(colour *);
  * If `ctrl' is null, _all_ controls in the dialog get refreshed
  * (for loading or saving entire sets of settings).
  */
-void dlg_refresh(control *);
+extern void dlg_refresh(control *);
 
 /*
  * Standard helper functions for reading a controlbox structure.
@@ -397,10 +411,10 @@ void dlg_refresh(control *);
  *          ... process this controlset ...
  *      }
  */
-int ctrl_find_path(controlbox *, char *path, int index);
+extern int ctrl_find_path(controlbox *, char * path, int index);
 
 /* Return the number of matching path elements at the starts of p1 and p2,
  * or INT_MAX if the paths are identical. */
-int ctrl_path_compare(char *p1, char *p2);
+extern int ctrl_path_compare(char * p1, char * p2);
 
 #endif
