@@ -1,4 +1,5 @@
 #include "minibidi.h"
+#include "term.h"  // UCSWIDE
 
 /************************************************************************
  * $Id: minibidi.c 6910 2006-11-18 15:10:48Z simon $
@@ -12,31 +13,25 @@
  *   http://www.unicode.org/reports/tr9/
  *
  * Author: Ahmad Khalifa
- *
- * -----------------
- * Revision Details:    (Updated by Revision Control System)
- * -----------------
- *  $Date: 2006-11-18 15:10:48 +0000 (Sat, 18 Nov 2006) $
- *  $Author: simon $
- *  $Revision: 6910 $
- *
  * (www.arabeyes.org - under MIT license)
  *
+ * Modified: Thomas Wolff:
+ * - extended bidi class detection to consider non-BMP characters
+ * - fixed mirroring of '('
+ * - fixed recognition of runs
+ * - sanitized handling of directional markers
+ *
  ************************************************************************/
 
 /*
  * TODO:
  * =====
- * - Explicit marks need to be handled (they are not 100% now)
- * - Ligatures
+ * - Explicit markers need to be handled (they are not 100% now),
+     esp. isolate markers
+ * - Ligatures (?)
  */
 
 
-#define LMASK	0x3F    /* Embedding Level mask */
-#define OMASK	0xC0    /* Override mask */
-#define OISL	0x80    /* Override is L */
-#define OISR	0x40    /* Override is R */
-
 /* Shaping Helpers */
 #define STYPE(xh) ((((xh) >= SHAPE_FIRST) && ((xh) <= SHAPE_LAST)) ? \
 shapetypes[(xh)-SHAPE_FIRST].type : SU) /*)) */
@@ -49,16 +44,11 @@ shapetypes[(xh)-SHAPE_FIRST].type : SU) /*)) */
 #define leastGreaterOdd(x) ( ((x)+1) | 1 )
 #define leastGreaterEven(x) ( ((x)+2) &~ 1 )
 
-/* character types */
-enum {
-  L, LRE, LRO, R, AL, RLE, RLO, PDF, EN, ES, ET, AN, CS, NSM, BN, B, S, WS, ON
-};
-
 /* Shaping Types */
 enum {
-  SL,   /* Left-Joining, doesnt exist in U+0600 - U+06FF */
-  SR,   /* Right-Joining, ie has Isolated, Final */
-  SD,   /* Dual-Joining, ie has Isolated, Final, Initial, Medial */
+  SL,   /* Left-Joining, doesn't exist in U+0600 - U+06FF */
+  SR,   /* Right-Joining, i.e. has Isolated, Final */
+  SD,   /* Dual-Joining, i.e. has Isolated, Final, Initial, Medial */
   SU,   /* Non-Joining */
   SC    /* Join-Causing, like U+0640 (TATWEEL) */
 };
@@ -119,14 +109,13 @@ static const shape_node shapetypes[] = {
 
 
 /*
- * Finds the index of a run with level equals tlevel
+ * Finds the index of a run with level tlevel or higher (!)
  */
 static int
-findIndexOfRun(uchar * level, int start, int count, int tlevel)
+find_run(uchar * levels, int start, int count, int tlevel)
 {
-  int i;
-  for (i = start; i < count; i++) {
-    if (tlevel == level[i]) {
+  for (int i = start; i < count; i++) {
+    if (tlevel <= levels[i]) {
       return i;
     }
   }
@@ -134,33 +123,29 @@ findIndexOfRun(uchar * level, int start, int count, int tlevel)
 }
 
 /*
- * Flips the text buffer, according to max level, and
- * all higher levels
+ * Flips runs in text buffer, of tlevel and all higher levels
  *
  * Input:
  * from: text buffer, on which to apply flipping
- * level: resolved levels buffer
- * max: the maximum level found in this line (should be uchar)
- * count: line size in bidi_char
+ * levels: resolved levels buffer
+ * tlevel: the level from which to flip runs
+ * count: line size in bidi_char and levels
  */
 static void
-flipThisRun(bidi_char * from, uchar * level, int max, int count)
+flip_runs(bidi_char * from, uchar * levels, int tlevel, int count)
 {
-  int i, j, k, tlevel;
-  bidi_char temp;
+  int i, j;
 
   j = i = 0;
   while (i < count && j < count) {
-
-   /* find the start of the run of level=max */
-    tlevel = max;
-    i = j = findIndexOfRun(level, i, count, max);
+   /* find the start of the next run */
+    i = j = find_run(levels, i, count, tlevel);
    /* find the end of the run */
-    while (i < count && tlevel <= level[i]) {
+    while (i < count && tlevel <= levels[i]) {
       i++;
     }
-    for (k = i - 1; k > j; k--, j++) {
-      temp = from[k];
+    for (int k = i - 1; k > j; k--, j++) {
+      bidi_char temp = from[k];
       from[k] = from[j];
       from[j] = temp;
     }
@@ -171,341 +156,16 @@ flipThisRun(bidi_char * from, uchar * level, int max, int count)
  * Returns the bidi character type of ch.
  *
  * The data table in this function is constructed from the Unicode
- * Character Database, downloadable from unicode.org at the URL
- * 
- *     http://www.unicode.org/Public/UNIDATA/UnicodeData.txt
- * 
- * by the following fragment of Perl:
- 
- perl -ne 'split ";"; $num = hex $_[0]; $type = $_[4];' \
- -e '$fl = ($_[1] =~ /First/ ? 1 : $_[1] =~ /Last/ ? 2 : 0);' \
- -e 'if ($type eq $runtype and ($runend == $num-1 or ' \
- -e '    ($fl==2 and $pfl==1))) {$runend = $num;} else { &reset; }' \
- -e '$pfl=$fl; END { &reset }; sub reset {' \
- -e 'printf"        {0x%04x, 0x%04x, %s},\n",$runstart,$runend,$runtype' \
- -e '  if defined $runstart and $runtype ne "ON";' \
- -e '$runstart=$runend=$num; $runtype=$type;}' \
- UnicodeData.txt
- 
+ * Character Database by the script mkbidiclasses
  */
-static uchar
-getType(wchar ch)
+uchar
+bidi_class(ucschar ch)
 {
   static const struct {
-    wchar first, last;
+    ucschar first, last;
     uchar type;
   } lookup[] = {
-    {0x0000, 0x0008, BN},  {0x0009, 0x0009, S},
-    {0x000a, 0x000a, B},   {0x000b, 0x000b, S},
-    {0x000c, 0x000c, WS},  {0x000d, 0x000d, B},
-    {0x000e, 0x001b, BN},  {0x001c, 0x001e, B},
-    {0x001f, 0x001f, S},   {0x0020, 0x0020, WS},
-    {0x0023, 0x0025, ET},  {0x002b, 0x002b, ES},
-    {0x002c, 0x002c, CS},  {0x002d, 0x002d, ES},
-    {0x002e, 0x002f, CS},  {0x0030, 0x0039, EN},
-    {0x003a, 0x003a, CS},  {0x0041, 0x005a, L},
-    {0x0061, 0x007a, L},   {0x007f, 0x0084, BN},
-    {0x0085, 0x0085, B},   {0x0086, 0x009f, BN},
-    {0x00a0, 0x00a0, CS},  {0x00a2, 0x00a5, ET},
-    {0x00aa, 0x00aa, L},   {0x00ad, 0x00ad, BN},
-    {0x00b0, 0x00b1, ET},  {0x00b2, 0x00b3, EN},
-    {0x00b5, 0x00b5, L},   {0x00b9, 0x00b9, EN},
-    {0x00ba, 0x00ba, L},   {0x00c0, 0x00d6, L},
-    {0x00d8, 0x00f6, L},   {0x00f8, 0x0236, L},
-    {0x0250, 0x02b8, L},   {0x02bb, 0x02c1, L},
-    {0x02d0, 0x02d1, L},   {0x02e0, 0x02e4, L},
-    {0x02ee, 0x02ee, L},   {0x0300, 0x0357, NSM},
-    {0x035d, 0x036f, NSM}, {0x037a, 0x037a, L},
-    {0x0386, 0x0386, L},   {0x0388, 0x038a, L},
-    {0x038c, 0x038c, L},   {0x038e, 0x03a1, L},
-    {0x03a3, 0x03ce, L},   {0x03d0, 0x03f5, L},
-    {0x03f7, 0x03fb, L},   {0x0400, 0x0482, L},
-    {0x0483, 0x0486, NSM}, {0x0488, 0x0489, NSM},
-    {0x048a, 0x04ce, L},   {0x04d0, 0x04f5, L},
-    {0x04f8, 0x04f9, L},   {0x0500, 0x050f, L},
-    {0x0531, 0x0556, L},   {0x0559, 0x055f, L},
-    {0x0561, 0x0587, L},   {0x0589, 0x0589, L},
-    {0x0591, 0x05a1, NSM}, {0x05a3, 0x05b9, NSM},
-    {0x05bb, 0x05bd, NSM}, {0x05be, 0x05be, R},
-    {0x05bf, 0x05bf, NSM}, {0x05c0, 0x05c0, R},
-    {0x05c1, 0x05c2, NSM}, {0x05c3, 0x05c3, R},
-    {0x05c4, 0x05c4, NSM}, {0x05d0, 0x05ea, R},
-    {0x05f0, 0x05f4, R},   {0x0600, 0x0603, AL},
-    {0x060c, 0x060c, CS},  {0x060d, 0x060d, AL},
-    {0x0610, 0x0615, NSM}, {0x061b, 0x061b, AL},
-    {0x061f, 0x061f, AL},  {0x0621, 0x063a, AL},
-    {0x0640, 0x064a, AL},  {0x064b, 0x0658, NSM},
-    {0x0660, 0x0669, AN},  {0x066a, 0x066a, ET},
-    {0x066b, 0x066c, AN},  {0x066d, 0x066f, AL},
-    {0x0670, 0x0670, NSM}, {0x0671, 0x06d5, AL},
-    {0x06d6, 0x06dc, NSM}, {0x06dd, 0x06dd, AL},
-    {0x06de, 0x06e4, NSM}, {0x06e5, 0x06e6, AL},
-    {0x06e7, 0x06e8, NSM}, {0x06ea, 0x06ed, NSM},
-    {0x06ee, 0x06ef, AL},  {0x06f0, 0x06f9, EN},
-    {0x06fa, 0x070d, AL},  {0x070f, 0x070f, BN},
-    {0x0710, 0x0710, AL},  {0x0711, 0x0711, NSM},
-    {0x0712, 0x072f, AL},  {0x0730, 0x074a, NSM},
-    {0x074d, 0x074f, AL},  {0x0780, 0x07a5, AL},
-    {0x07a6, 0x07b0, NSM}, {0x07b1, 0x07b1, AL},
-    {0x0901, 0x0902, NSM}, {0x0903, 0x0939, L},
-    {0x093c, 0x093c, NSM}, {0x093d, 0x0940, L},
-    {0x0941, 0x0948, NSM}, {0x0949, 0x094c, L},
-    {0x094d, 0x094d, NSM}, {0x0950, 0x0950, L},
-    {0x0951, 0x0954, NSM}, {0x0958, 0x0961, L},
-    {0x0962, 0x0963, NSM}, {0x0964, 0x0970, L},
-    {0x0981, 0x0981, NSM}, {0x0982, 0x0983, L},
-    {0x0985, 0x098c, L},   {0x098f, 0x0990, L},
-    {0x0993, 0x09a8, L},   {0x09aa, 0x09b0, L},
-    {0x09b2, 0x09b2, L},   {0x09b6, 0x09b9, L},
-    {0x09bc, 0x09bc, NSM}, {0x09bd, 0x09c0, L},
-    {0x09c1, 0x09c4, NSM}, {0x09c7, 0x09c8, L},
-    {0x09cb, 0x09cc, L},   {0x09cd, 0x09cd, NSM},
-    {0x09d7, 0x09d7, L},   {0x09dc, 0x09dd, L},
-    {0x09df, 0x09e1, L},   {0x09e2, 0x09e3, NSM},
-    {0x09e6, 0x09f1, L},   {0x09f2, 0x09f3, ET},
-    {0x09f4, 0x09fa, L},   {0x0a01, 0x0a02, NSM},
-    {0x0a03, 0x0a03, L},   {0x0a05, 0x0a0a, L},
-    {0x0a0f, 0x0a10, L},   {0x0a13, 0x0a28, L},
-    {0x0a2a, 0x0a30, L},   {0x0a32, 0x0a33, L},
-    {0x0a35, 0x0a36, L},   {0x0a38, 0x0a39, L},
-    {0x0a3c, 0x0a3c, NSM}, {0x0a3e, 0x0a40, L},
-    {0x0a41, 0x0a42, NSM}, {0x0a47, 0x0a48, NSM},
-    {0x0a4b, 0x0a4d, NSM}, {0x0a59, 0x0a5c, L},
-    {0x0a5e, 0x0a5e, L},   {0x0a66, 0x0a6f, L},
-    {0x0a70, 0x0a71, NSM}, {0x0a72, 0x0a74, L},
-    {0x0a81, 0x0a82, NSM}, {0x0a83, 0x0a83, L},
-    {0x0a85, 0x0a8d, L},   {0x0a8f, 0x0a91, L},
-    {0x0a93, 0x0aa8, L},   {0x0aaa, 0x0ab0, L},
-    {0x0ab2, 0x0ab3, L},   {0x0ab5, 0x0ab9, L},
-    {0x0abc, 0x0abc, NSM}, {0x0abd, 0x0ac0, L},
-    {0x0ac1, 0x0ac5, NSM}, {0x0ac7, 0x0ac8, NSM},
-    {0x0ac9, 0x0ac9, L},   {0x0acb, 0x0acc, L},
-    {0x0acd, 0x0acd, NSM}, {0x0ad0, 0x0ad0, L},
-    {0x0ae0, 0x0ae1, L},   {0x0ae2, 0x0ae3, NSM},
-    {0x0ae6, 0x0aef, L},   {0x0af1, 0x0af1, ET},
-    {0x0b01, 0x0b01, NSM}, {0x0b02, 0x0b03, L},
-    {0x0b05, 0x0b0c, L},   {0x0b0f, 0x0b10, L},
-    {0x0b13, 0x0b28, L},   {0x0b2a, 0x0b30, L},
-    {0x0b32, 0x0b33, L},   {0x0b35, 0x0b39, L},
-    {0x0b3c, 0x0b3c, NSM}, {0x0b3d, 0x0b3e, L},
-    {0x0b3f, 0x0b3f, NSM}, {0x0b40, 0x0b40, L},
-    {0x0b41, 0x0b43, NSM}, {0x0b47, 0x0b48, L},
-    {0x0b4b, 0x0b4c, L},   {0x0b4d, 0x0b4d, NSM},
-    {0x0b56, 0x0b56, NSM}, {0x0b57, 0x0b57, L},
-    {0x0b5c, 0x0b5d, L},   {0x0b5f, 0x0b61, L},
-    {0x0b66, 0x0b71, L},   {0x0b82, 0x0b82, NSM},
-    {0x0b83, 0x0b83, L},   {0x0b85, 0x0b8a, L},
-    {0x0b8e, 0x0b90, L},   {0x0b92, 0x0b95, L},
-    {0x0b99, 0x0b9a, L},   {0x0b9c, 0x0b9c, L},
-    {0x0b9e, 0x0b9f, L},   {0x0ba3, 0x0ba4, L},
-    {0x0ba8, 0x0baa, L},   {0x0bae, 0x0bb5, L},
-    {0x0bb7, 0x0bb9, L},   {0x0bbe, 0x0bbf, L},
-    {0x0bc0, 0x0bc0, NSM}, {0x0bc1, 0x0bc2, L},
-    {0x0bc6, 0x0bc8, L},   {0x0bca, 0x0bcc, L},
-    {0x0bcd, 0x0bcd, NSM}, {0x0bd7, 0x0bd7, L},
-    {0x0be7, 0x0bf2, L},   {0x0bf9, 0x0bf9, ET},
-    {0x0c01, 0x0c03, L},   {0x0c05, 0x0c0c, L},
-    {0x0c0e, 0x0c10, L},   {0x0c12, 0x0c28, L},
-    {0x0c2a, 0x0c33, L},   {0x0c35, 0x0c39, L},
-    {0x0c3e, 0x0c40, NSM}, {0x0c41, 0x0c44, L},
-    {0x0c46, 0x0c48, NSM}, {0x0c4a, 0x0c4d, NSM},
-    {0x0c55, 0x0c56, NSM}, {0x0c60, 0x0c61, L},
-    {0x0c66, 0x0c6f, L},   {0x0c82, 0x0c83, L},
-    {0x0c85, 0x0c8c, L},   {0x0c8e, 0x0c90, L},
-    {0x0c92, 0x0ca8, L},   {0x0caa, 0x0cb3, L},
-    {0x0cb5, 0x0cb9, L},   {0x0cbc, 0x0cbc, NSM},
-    {0x0cbd, 0x0cc4, L},   {0x0cc6, 0x0cc8, L},
-    {0x0cca, 0x0ccb, L},   {0x0ccc, 0x0ccd, NSM},
-    {0x0cd5, 0x0cd6, L},   {0x0cde, 0x0cde, L},
-    {0x0ce0, 0x0ce1, L},   {0x0ce6, 0x0cef, L},
-    {0x0d02, 0x0d03, L},   {0x0d05, 0x0d0c, L},
-    {0x0d0e, 0x0d10, L},   {0x0d12, 0x0d28, L},
-    {0x0d2a, 0x0d39, L},   {0x0d3e, 0x0d40, L},
-    {0x0d41, 0x0d43, NSM}, {0x0d46, 0x0d48, L},
-    {0x0d4a, 0x0d4c, L},   {0x0d4d, 0x0d4d, NSM},
-    {0x0d57, 0x0d57, L},   {0x0d60, 0x0d61, L},
-    {0x0d66, 0x0d6f, L},   {0x0d82, 0x0d83, L},
-    {0x0d85, 0x0d96, L},   {0x0d9a, 0x0db1, L},
-    {0x0db3, 0x0dbb, L},   {0x0dbd, 0x0dbd, L},
-    {0x0dc0, 0x0dc6, L},   {0x0dca, 0x0dca, NSM},
-    {0x0dcf, 0x0dd1, L},   {0x0dd2, 0x0dd4, NSM},
-    {0x0dd6, 0x0dd6, NSM}, {0x0dd8, 0x0ddf, L},
-    {0x0df2, 0x0df4, L},   {0x0e01, 0x0e30, L},
-    {0x0e31, 0x0e31, NSM}, {0x0e32, 0x0e33, L},
-    {0x0e34, 0x0e3a, NSM}, {0x0e3f, 0x0e3f, ET},
-    {0x0e40, 0x0e46, L},   {0x0e47, 0x0e4e, NSM},
-    {0x0e4f, 0x0e5b, L},   {0x0e81, 0x0e82, L},
-    {0x0e84, 0x0e84, L},   {0x0e87, 0x0e88, L},
-    {0x0e8a, 0x0e8a, L},   {0x0e8d, 0x0e8d, L},
-    {0x0e94, 0x0e97, L},   {0x0e99, 0x0e9f, L},
-    {0x0ea1, 0x0ea3, L},   {0x0ea5, 0x0ea5, L},
-    {0x0ea7, 0x0ea7, L},   {0x0eaa, 0x0eab, L},
-    {0x0ead, 0x0eb0, L},   {0x0eb1, 0x0eb1, NSM},
-    {0x0eb2, 0x0eb3, L},   {0x0eb4, 0x0eb9, NSM},
-    {0x0ebb, 0x0ebc, NSM}, {0x0ebd, 0x0ebd, L},
-    {0x0ec0, 0x0ec4, L},   {0x0ec6, 0x0ec6, L},
-    {0x0ec8, 0x0ecd, NSM}, {0x0ed0, 0x0ed9, L},
-    {0x0edc, 0x0edd, L},   {0x0f00, 0x0f17, L},
-    {0x0f18, 0x0f19, NSM}, {0x0f1a, 0x0f34, L},
-    {0x0f35, 0x0f35, NSM}, {0x0f36, 0x0f36, L},
-    {0x0f37, 0x0f37, NSM}, {0x0f38, 0x0f38, L},
-    {0x0f39, 0x0f39, NSM}, {0x0f3e, 0x0f47, L},
-    {0x0f49, 0x0f6a, L},   {0x0f71, 0x0f7e, NSM},
-    {0x0f7f, 0x0f7f, L},   {0x0f80, 0x0f84, NSM},
-    {0x0f85, 0x0f85, L},   {0x0f86, 0x0f87, NSM},
-    {0x0f88, 0x0f8b, L},   {0x0f90, 0x0f97, NSM},
-    {0x0f99, 0x0fbc, NSM}, {0x0fbe, 0x0fc5, L},
-    {0x0fc6, 0x0fc6, NSM}, {0x0fc7, 0x0fcc, L},
-    {0x0fcf, 0x0fcf, L},   {0x1000, 0x1021, L},
-    {0x1023, 0x1027, L},   {0x1029, 0x102a, L},
-    {0x102c, 0x102c, L},   {0x102d, 0x1030, NSM},
-    {0x1031, 0x1031, L},   {0x1032, 0x1032, NSM},
-    {0x1036, 0x1037, NSM}, {0x1038, 0x1038, L},
-    {0x1039, 0x1039, NSM}, {0x1040, 0x1057, L},
-    {0x1058, 0x1059, NSM}, {0x10a0, 0x10c5, L},
-    {0x10d0, 0x10f8, L},   {0x10fb, 0x10fb, L},
-    {0x1100, 0x1159, L},   {0x115f, 0x11a2, L},
-    {0x11a8, 0x11f9, L},   {0x1200, 0x1206, L},
-    {0x1208, 0x1246, L},   {0x1248, 0x1248, L},
-    {0x124a, 0x124d, L},   {0x1250, 0x1256, L},
-    {0x1258, 0x1258, L},   {0x125a, 0x125d, L},
-    {0x1260, 0x1286, L},   {0x1288, 0x1288, L},
-    {0x128a, 0x128d, L},   {0x1290, 0x12ae, L},
-    {0x12b0, 0x12b0, L},   {0x12b2, 0x12b5, L},
-    {0x12b8, 0x12be, L},   {0x12c0, 0x12c0, L},
-    {0x12c2, 0x12c5, L},   {0x12c8, 0x12ce, L},
-    {0x12d0, 0x12d6, L},   {0x12d8, 0x12ee, L},
-    {0x12f0, 0x130e, L},   {0x1310, 0x1310, L},
-    {0x1312, 0x1315, L},   {0x1318, 0x131e, L},
-    {0x1320, 0x1346, L},   {0x1348, 0x135a, L},
-    {0x1361, 0x137c, L},   {0x13a0, 0x13f4, L},
-    {0x1401, 0x1676, L},   {0x1680, 0x1680, WS},
-    {0x1681, 0x169a, L},   {0x16a0, 0x16f0, L},
-    {0x1700, 0x170c, L},   {0x170e, 0x1711, L},
-    {0x1712, 0x1714, NSM}, {0x1720, 0x1731, L},
-    {0x1732, 0x1734, NSM}, {0x1735, 0x1736, L},
-    {0x1740, 0x1751, L},   {0x1752, 0x1753, NSM},
-    {0x1760, 0x176c, L},   {0x176e, 0x1770, L},
-    {0x1772, 0x1773, NSM}, {0x1780, 0x17b6, L},
-    {0x17b7, 0x17bd, NSM}, {0x17be, 0x17c5, L},
-    {0x17c6, 0x17c6, NSM}, {0x17c7, 0x17c8, L},
-    {0x17c9, 0x17d3, NSM}, {0x17d4, 0x17da, L},
-    {0x17db, 0x17db, ET},  {0x17dc, 0x17dc, L},
-    {0x17dd, 0x17dd, NSM}, {0x17e0, 0x17e9, L},
-    {0x180b, 0x180d, NSM}, {0x180e, 0x180e, WS},
-    {0x1810, 0x1819, L},   {0x1820, 0x1877, L},
-    {0x1880, 0x18a8, L},   {0x18a9, 0x18a9, NSM},
-    {0x1900, 0x191c, L},   {0x1920, 0x1922, NSM},
-    {0x1923, 0x1926, L},   {0x1927, 0x192b, NSM},
-    {0x1930, 0x1931, L},   {0x1932, 0x1932, NSM},
-    {0x1933, 0x1938, L},   {0x1939, 0x193b, NSM},
-    {0x1946, 0x196d, L},   {0x1970, 0x1974, L},
-    {0x1d00, 0x1d6b, L},   {0x1e00, 0x1e9b, L},
-    {0x1ea0, 0x1ef9, L},   {0x1f00, 0x1f15, L},
-    {0x1f18, 0x1f1d, L},   {0x1f20, 0x1f45, L},
-    {0x1f48, 0x1f4d, L},   {0x1f50, 0x1f57, L},
-    {0x1f59, 0x1f59, L},   {0x1f5b, 0x1f5b, L},
-    {0x1f5d, 0x1f5d, L},   {0x1f5f, 0x1f7d, L},
-    {0x1f80, 0x1fb4, L},   {0x1fb6, 0x1fbc, L},
-    {0x1fbe, 0x1fbe, L},   {0x1fc2, 0x1fc4, L},
-    {0x1fc6, 0x1fcc, L},   {0x1fd0, 0x1fd3, L},
-    {0x1fd6, 0x1fdb, L},   {0x1fe0, 0x1fec, L},
-    {0x1ff2, 0x1ff4, L},   {0x1ff6, 0x1ffc, L},
-    {0x2000, 0x200a, WS},  {0x200b, 0x200d, BN},
-    {0x200e, 0x200e, L},   {0x200f, 0x200f, R},
-    {0x2028, 0x2028, WS},  {0x2029, 0x2029, B},
-    {0x202a, 0x202a, LRE}, {0x202b, 0x202b, RLE},
-    {0x202c, 0x202c, PDF}, {0x202d, 0x202d, LRO},
-    {0x202e, 0x202e, RLO}, {0x202f, 0x202f, WS},
-    {0x2030, 0x2034, ET},  {0x2044, 0x2044, CS},
-    {0x205f, 0x205f, WS},  {0x2060, 0x2063, BN},
-    {0x206a, 0x206f, BN},  {0x2070, 0x2070, EN},
-    {0x2071, 0x2071, L},   {0x2074, 0x2079, EN},
-    {0x207a, 0x207b, ET},  {0x207f, 0x207f, L},
-    {0x2080, 0x2089, EN},  {0x208a, 0x208b, ET},
-    {0x20a0, 0x20b1, ET},  {0x20d0, 0x20ea, NSM},
-    {0x2102, 0x2102, L},   {0x2107, 0x2107, L},
-    {0x210a, 0x2113, L},   {0x2115, 0x2115, L},
-    {0x2119, 0x211d, L},   {0x2124, 0x2124, L},
-    {0x2126, 0x2126, L},   {0x2128, 0x2128, L},
-    {0x212a, 0x212d, L},   {0x212e, 0x212e, ET},
-    {0x212f, 0x2131, L},   {0x2133, 0x2139, L},
-    {0x213d, 0x213f, L},   {0x2145, 0x2149, L},
-    {0x2160, 0x2183, L},   {0x2212, 0x2213, ET},
-    {0x2336, 0x237a, L},   {0x2395, 0x2395, L},
-    {0x2488, 0x249b, EN},  {0x249c, 0x24e9, L},
-    {0x2800, 0x28ff, L},   {0x3000, 0x3000, WS},
-    {0x3005, 0x3007, L},   {0x3021, 0x3029, L},
-    {0x302a, 0x302f, NSM}, {0x3031, 0x3035, L},
-    {0x3038, 0x303c, L},   {0x3041, 0x3096, L},
-    {0x3099, 0x309a, NSM}, {0x309d, 0x309f, L},
-    {0x30a1, 0x30fa, L},   {0x30fc, 0x30ff, L},
-    {0x3105, 0x312c, L},   {0x3131, 0x318e, L},
-    {0x3190, 0x31b7, L},   {0x31f0, 0x321c, L},
-    {0x3220, 0x3243, L},   {0x3260, 0x327b, L},
-    {0x327f, 0x32b0, L},   {0x32c0, 0x32cb, L},
-    {0x32d0, 0x32fe, L},   {0x3300, 0x3376, L},
-    {0x337b, 0x33dd, L},   {0x33e0, 0x33fe, L},
-    {0x3400, 0x4db5, L},   {0x4e00, 0x9fa5, L},
-    {0xa000, 0xa48c, L},   {0xac00, 0xd7a3, L},
-    {0xd800, 0xfa2d, L},   {0xfa30, 0xfa6a, L},
-    {0xfb00, 0xfb06, L},   {0xfb13, 0xfb17, L},
-    {0xfb1d, 0xfb1d, R},   {0xfb1e, 0xfb1e, NSM},
-    {0xfb1f, 0xfb28, R},   {0xfb29, 0xfb29, ET},
-    {0xfb2a, 0xfb36, R},   {0xfb38, 0xfb3c, R},
-    {0xfb3e, 0xfb3e, R},   {0xfb40, 0xfb41, R},
-    {0xfb43, 0xfb44, R},   {0xfb46, 0xfb4f, R},
-    {0xfb50, 0xfbb1, AL},  {0xfbd3, 0xfd3d, AL},
-    {0xfd50, 0xfd8f, AL},  {0xfd92, 0xfdc7, AL},
-    {0xfdf0, 0xfdfc, AL},  {0xfe00, 0xfe0f, NSM},
-    {0xfe20, 0xfe23, NSM}, {0xfe50, 0xfe50, CS},
-    {0xfe52, 0xfe52, CS},  {0xfe55, 0xfe55, CS},
-    {0xfe5f, 0xfe5f, ET},  {0xfe62, 0xfe63, ET},
-    {0xfe69, 0xfe6a, ET},  {0xfe70, 0xfe74, AL},
-    {0xfe76, 0xfefc, AL},  {0xfeff, 0xfeff, BN},
-    {0xff03, 0xff05, ET},  {0xff0b, 0xff0b, ET},
-    {0xff0c, 0xff0c, CS},  {0xff0d, 0xff0d, ET},
-    {0xff0e, 0xff0e, CS},  {0xff0f, 0xff0f, ES},
-    {0xff10, 0xff19, EN},  {0xff1a, 0xff1a, CS},
-    {0xff21, 0xff3a, L},   {0xff41, 0xff5a, L},
-    {0xff66, 0xffbe, L},   {0xffc2, 0xffc7, L},
-    {0xffca, 0xffcf, L},   {0xffd2, 0xffd7, L},
-    {0xffda, 0xffdc, L},   {0xffe0, 0xffe1, ET},
-    {0xffe5, 0xffe6, ET}
-    /*
-    {0x10000, 0x1000b, L},   {0x1000d, 0x10026, L},
-    {0x10028, 0x1003a, L},   {0x1003c, 0x1003d, L},
-    {0x1003f, 0x1004d, L},   {0x10050, 0x1005d, L},
-    {0x10080, 0x100fa, L},   {0x10100, 0x10100, L},
-    {0x10102, 0x10102, L},   {0x10107, 0x10133, L},
-    {0x10137, 0x1013f, L},   {0x10300, 0x1031e, L},
-    {0x10320, 0x10323, L},   {0x10330, 0x1034a, L},
-    {0x10380, 0x1039d, L},   {0x1039f, 0x1039f, L},
-    {0x10400, 0x1049d, L},   {0x104a0, 0x104a9, L},
-    {0x10800, 0x10805, R},   {0x10808, 0x10808, R},
-    {0x1080a, 0x10835, R},   {0x10837, 0x10838, R},
-    {0x1083c, 0x1083c, R},   {0x1083f, 0x1083f, R},
-    {0x1d000, 0x1d0f5, L},   {0x1d100, 0x1d126, L},
-    {0x1d12a, 0x1d166, L},   {0x1d167, 0x1d169, NSM},
-    {0x1d16a, 0x1d172, L},   {0x1d173, 0x1d17a, BN},
-    {0x1d17b, 0x1d182, NSM}, {0x1d183, 0x1d184, L},
-    {0x1d185, 0x1d18b, NSM}, {0x1d18c, 0x1d1a9, L},
-    {0x1d1aa, 0x1d1ad, NSM}, {0x1d1ae, 0x1d1dd, L},
-    {0x1d400, 0x1d454, L},   {0x1d456, 0x1d49c, L},
-    {0x1d49e, 0x1d49f, L},   {0x1d4a2, 0x1d4a2, L},
-    {0x1d4a5, 0x1d4a6, L},   {0x1d4a9, 0x1d4ac, L},
-    {0x1d4ae, 0x1d4b9, L},   {0x1d4bb, 0x1d4bb, L},
-    {0x1d4bd, 0x1d4c3, L},   {0x1d4c5, 0x1d505, L},
-    {0x1d507, 0x1d50a, L},   {0x1d50d, 0x1d514, L},
-    {0x1d516, 0x1d51c, L},   {0x1d51e, 0x1d539, L},
-    {0x1d53b, 0x1d53e, L},   {0x1d540, 0x1d544, L},
-    {0x1d546, 0x1d546, L},   {0x1d54a, 0x1d550, L},
-    {0x1d552, 0x1d6a3, L},   {0x1d6a8, 0x1d7c9, L},
-    {0x1d7ce, 0x1d7ff, EN},  {0x20000, 0x2a6d6, L},
-    {0x2f800, 0x2fa1d, L},   {0xe0001, 0xe0001, BN},
-    {0xe0020, 0xe007f, BN},  {0xe0100, 0xe01ef, NSM},
-    {0xf0000, 0xffffd, L},   {0x100000, 0x10fffd, L}
-*/
+#include "bidiclasses.t"
   };
 
   int i, j, k;
@@ -553,8 +213,9 @@ getType(wchar ch)
  * important because all such characters are right-to-left so it
  * would have flagged them anyway.)
  */
+//#define is_rtl(c)	is_rtl_class(bidi_class(c))
 bool
-is_rtl(wchar c)
+is_rtl_class(uchar bc)
 {
  /*
   * After careful reading of the Unicode bidi algorithm (URL as
@@ -564,52 +225,38 @@ is_rtl(wchar c)
   * character in any of those classes will be displayed
   * uniformly left-to-right by the Unicode bidi algorithm.
   */
-  const int mask = (1 << R) | (1 << AL) | (1 << RLE) | (1 << RLO);
+  const int mask = (1 << R) | (1 << AL) | (1 << RLE) | (1 << RLO)
+                 | (1 << RLI) | (1 << FSI)
+                 ;
 
-  return mask & (1 << (getType(c)));
+  return mask & (1 << (bc));
 }
 
-/*
- * The most significant 2 bits of each level are used to store
- * Override status of each character
- * This function sets the override bits of level according
- * to the value in override, and reurns the new byte.
- */
-static uchar
-setOverrideBits(uchar level, uchar override)
+bool
+is_sep_class(uchar bc)
 {
-  if (override == ON)
-    return level;
-  else if (override == R)
-    return level | OISR;
-  else if (override == L)
-    return level | OISL;
-  return level;
+  const int mask = (1 << B) | (1 << S) | (1 << BN) | (1 << WS) | (1 << ON);
+
+  return mask & (1 << (bc));
 }
 
-/*
- * Find the most recent run of the same value in `level', and
- * return the value _before_ it. Used to process U+202C POP
- * DIRECTIONAL FORMATTING.
- */
-static int
-getPreviousLevel(uchar * level, int from)
+bool
+is_punct_class(uchar bc)
 {
-  if (from > 0) {
-    uchar current = level[--from];
+  const int mask = (1 << BN) | (1 << CS) | (1 << EN) | (1 << ES) | (1 << ET);
 
-    while (from >= 0 && level[from] == current)
-      from--;
-
-    if (from >= 0)
-      return level[from];
-
-    return -1;
-  }
-  else
-    return -1;
+  return mask & (1 << (bc));
 }
 
+static bool
+is_NI(uchar bc)
+{
+  const int mask = (1 << B) | (1 <<  S) | (1 <<  WS) | (1 <<  ON) | (1 <<  FSI) | (1 <<  LRI) | (1 <<  RLI) | (1 <<  PDI);
+
+  return mask & (1 << (bc));
+}
+
+
 /* The Main shaping function, and the only one to be used
  * by the outside world.
  *
@@ -690,8 +337,8 @@ do_shape(bidi_char * line, bidi_char * to, int count)
   return 1;
 }
 
-static wchar
-mirror(wchar c)
+static ucschar
+mirror(ucschar c)
 {
   static const struct { wchar from, to; } pairs[] = {
     {0x0028, 0x0029}, {0x0029, 0x0028}, {0x003C, 0x003E}, {0x003E, 0x003C},
@@ -775,7 +422,7 @@ mirror(wchar c)
     {0xFF5B, 0xFF5D}, {0xFF5D, 0xFF5B}, {0xFF5F, 0xFF60}, {0xFF60, 0xFF5F},
     {0xFF62, 0xFF63}, {0xFF63, 0xFF62}
   };
-  int i = 0;
+  int i = -1;
   int j = lengthof(pairs);
   while (j - i > 1) {
     int k = (i + j) / 2;
@@ -804,13 +451,22 @@ do_bidi(bidi_char * line, int count)
   uchar currentEmbedding;
   uchar currentOverride;
   uchar tempType;
-  int i, j, yes, bover;
+  int i, j, yes;
+
+  uchar bidi_class_of(int i) {
+    if (i && line[i].wc == UCSWIDE)
+      i--;
+    return bidi_class(line[i].wc);
+  }
 
  /* Check the presence of R or AL types as optimization */
   yes = 0;
   for (i = 0; i < count; i++) {
-    int type = getType(line[i].wc);
-    if (type == R || type == AL) {
+    int type = bidi_class_of(i);
+    if (type == R || type == AL
+        || type == RLE || type == LRE || type == RLO || type == LRO || type == PDF
+        || type == LRI || type == RLI || type == FSI || type == PDI
+       ) {
       yes = 1;
       break;
     }
@@ -829,27 +485,103 @@ do_bidi(bidi_char * line, int count)
   */
 
  /* Rule (P2), (P3)
-  * P2. In each paragraph, find the first character of type L, AL, or R.
+  * P2. In each paragraph, find the first character of type L, AL, or R 
+    while skipping over any characters between an isolate initiator and 
+    its matching PDI or, if it has no matching PDI, the end of the paragraph.
   * P3. If a character is found in P2 and it is of type AL or R, then set
   * the paragraph embedding level to one; otherwise, set it to zero.
   */
   paragraphLevel = 0;
+  int isolateLevel = 0;
   for (i = 0; i < count; i++) {
-    int type = getType(line[i].wc);
-    if (type == R || type == AL) {
-      paragraphLevel = 1;
-      break;
+    int type = bidi_class_of(i);
+    if (type == LRI || type == RLI || type == FSI)
+      isolateLevel++;
+    else if (type == PDI)
+      isolateLevel--;
+    else if (isolateLevel == 0) {
+      if (type == R || type == AL) {
+        paragraphLevel = 1;
+        break;
+      }
+      else if (type == L)
+        break;
     }
-    else if (type == L)
-      break;
   }
 
  /* Rule (X1)
-  * X1. Begin by setting the current embedding level to the paragraph
-  * embedding level. Set the directional override status to neutral.
+    X1. At the beginning of a paragraph, perform the following steps:
+  • Set the stack to empty.
+  • Push onto the stack an entry consisting of the paragraph embedding level,
+    a neutral directional override status, and a false directional isolate status.
+  • Set the overflow isolate count to zero.
+  • Set the overflow embedding count to zero.
+  • Set the valid isolate count to zero.
+  • Process each character iteratively, applying rules X2 through X8. 
+    Only embedding levels from 0 through max_depth are valid in this phase. 
+    (Note that in the resolution of levels in rules I1 and I2, 
+    the maximum embedding level of max_depth+1 can be reached.)
   */
   currentEmbedding = paragraphLevel;
   currentOverride = ON;
+  bool currentIsolate = false;
+
+  // By making the dss as large as the whole line, we avoid overflow handling.
+  uchar dss_emb[count + 1];
+  uchar dss_ovr[count + 1];
+  bool dss_isol[count + 1];
+  int dss_top = -1;
+
+  int countdss() { return dss_top + 1; }
+
+  void pushdss() {
+    dss_top++;
+    dss_emb[dss_top] = currentEmbedding;
+    dss_ovr[dss_top] = currentOverride;
+    dss_isol[dss_top] = currentIsolate;
+  }
+
+  void popdss() {
+    // remove top
+    if (dss_top >= 0)
+      dss_top--;
+    // then set current values to new top
+    if (dss_top >= 0) {
+      currentEmbedding = dss_emb[dss_top];
+      currentOverride = dss_ovr[dss_top];
+      currentIsolate = dss_isol[dss_top];
+    }
+  }
+
+  pushdss();
+  //int ovfIsolate = 0;
+  //int ovfEmbedding = 0;
+  isolateLevel = 0;
+
+#define dont_debug_bidi
+
+#ifdef debug_bidi
+  void trace_bidi(char * tag) {
+    if ((*line).wc == 'x') {
+      printf("%s ", tag);
+      for (int i = 0; i < count; i++) {
+        if (line[i].wc == ' ')
+          printf(" ");
+        else if (line[i].wc < 0x80)
+          printf("%d:%c/%d ", levels[i], line[i].wc, types[i]);
+        else
+          printf("%d:%X/%d ", levels[i], line[i].wc, types[i]);
+      }
+      printf("\n");
+    }
+  }
+  void trace_mark(char * tag) {
+    (void)tag;
+  }
+#else
+#define trace_bidi(tag)	
+#define trace_mark(tag)	
+#endif
 
  /* Rule (X2), (X3), (X4), (X5), (X6), (X7), (X8)
   * X2. With each RLE, compute the least greater odd embedding level.
@@ -859,9 +591,9 @@ do_bidi(bidi_char * line, int count)
   * X6. For all types besides RLE, LRE, RLO, LRO, and PDF:
   *          a. Set the level of the current character to the current
   *              embedding level.
-  *          b.  Whenever the directional override status is not neutral,
-  *               reset the current character type to the directional
-  *               override status.
+  *          b. Whenever the directional override status is not neutral,
+  *              reset the current character type to the directional
+  *              override status.
   * X7. With each PDF, determine the matching embedding or override code.
   * If there was a valid matching code, restore (pop) the last
   * remembered (pushed) embedding level and directional override.
@@ -869,55 +601,102 @@ do_bidi(bidi_char * line, int count)
   * terminated at the end of each paragraph. Paragraph separators are not
   * included in the embedding. (Useless here) NOT IMPLEMENTED
   */
-  bover = 0;
   for (i = 0; i < count; i++) {
-    tempType = getType(line[i].wc);
+    tempType = bidi_class_of(i);
+    levels[i] = currentEmbedding;
+
+    if (tempType == FSI) {
+      int lvl = 0;
+      tempType = LRI;
+      for (int k = i + 1; k < count; k++) {
+        uchar kType = bidi_class_of(k);
+        if (kType == FSI || kType == RLI || kType == LRI)
+          lvl++;
+        else if (kType == PDI) {
+          if (lvl)
+            lvl--;
+          else
+            break;
+        }
+        else if (kType == R || kType == AL) {
+          tempType = RLI;
+          break;
+        }
+        else if (kType == L)
+          break;
+      }
+    }
     switch (tempType) {
       when RLE:
-        currentEmbedding = levels[i] = leastGreaterOdd(currentEmbedding);
-        levels[i] = setOverrideBits(levels[i], currentOverride);
+        currentEmbedding = leastGreaterOdd(currentEmbedding);
         currentOverride = ON;
+        currentIsolate = false;
+        pushdss();
+        trace_mark("RLE");
       when LRE:
-        currentEmbedding = levels[i] = leastGreaterEven(currentEmbedding);
-        levels[i] = setOverrideBits(levels[i], currentOverride);
+        currentEmbedding = leastGreaterEven(currentEmbedding);
         currentOverride = ON;
+        currentIsolate = false;
+        pushdss();
+        trace_mark("LRE");
       when RLO:
-        currentEmbedding = levels[i] = leastGreaterOdd(currentEmbedding);
-        tempType = currentOverride = R;
-        bover = 1;
+        currentEmbedding = leastGreaterOdd(currentEmbedding);
+        currentOverride = R;
+        currentIsolate = false;
+        pushdss();
+        trace_mark("RLO");
       when LRO:
-        currentEmbedding = levels[i] = leastGreaterEven(currentEmbedding);
-        tempType = currentOverride = L;
-        bover = 1;
+        currentEmbedding = leastGreaterEven(currentEmbedding);
+        currentOverride = L;
+        currentIsolate = false;
+        pushdss();
+        trace_mark("LRO");
+      when RLI:
+        if (currentOverride != ON)
+          tempType = currentOverride;
+        currentEmbedding = leastGreaterOdd(currentEmbedding);
+        isolateLevel++;
+        currentOverride = ON;
+        currentIsolate = true;
+        pushdss();
+        trace_mark("RLI");
+      when LRI:
+        if (currentOverride != ON)
+          tempType = currentOverride;
+        currentEmbedding = leastGreaterEven(currentEmbedding);
+        isolateLevel++;
+        currentOverride = ON;
+        currentIsolate = true;
+        pushdss();
+        trace_mark("LRI");
       when PDF:
-        if (getPreviousLevel(levels, i) == -1) {
-          currentEmbedding = paragraphLevel;
-          currentOverride = ON;
-        }
-        else {
-          currentOverride = currentEmbedding & OMASK;
-          currentEmbedding = currentEmbedding & ~OMASK;
-        }
+        if (!currentIsolate && countdss() >= 2)
+          popdss();
         levels[i] = currentEmbedding;
-       /* Whitespace is treated as neutral for now */
-      when WS or S:
+        trace_mark("PDF");
+      when PDI:
+        if (isolateLevel) {
+          while (!currentIsolate && countdss() > 0)
+            popdss();
+          popdss();
+          isolateLevel--;
+        }
+        if (currentOverride != ON)
+          tempType = currentOverride;
         levels[i] = currentEmbedding;
+        trace_mark("PDI");
+      when WS or S: /* Whitespace is treated as neutral for now */
         tempType = ON;
         if (currentOverride != ON)
           tempType = currentOverride;
       otherwise:
-        levels[i] = currentEmbedding;
         if (currentOverride != ON)
           tempType = currentOverride;
     }
     types[i] = tempType;
   }
- /* this clears out all overrides, so we can use levels safely... */
- /* checks bover first */
-  if (bover)
-    for (i = 0; i < count; i++)
-      levels[i] = levels[i] & LMASK;
 
+  trace_bidi("<X9");
  /* Rule (X9)
   * X9. Remove all RLE, LRE, RLO, LRO, PDF, and BN codes.
   * Here, they're converted to BN.
@@ -929,25 +708,40 @@ do_bidi(bidi_char * line, int count)
     }
   }
 
+ /* Rule (X10) NOT IMPLEMENTED
+  * X10. Handle isolating run sequences...
+  */
+
  /* Rule (W1)
   * W1. Examine each non-spacing mark (NSM) in the level run, and change
   * the type of the NSM to the type of the previous character. If the NSM
   * is at the start of the level run, it will get the type of sor.
+  // TODO: check
+    W1. Examine each nonspacing mark (NSM) in the isolating run sequence, 
+    and change the type of the NSM 
+    to Other Neutral if the previous character is an isolate initiator or PDI, 
+    and to the type of the previous character otherwise. 
+    If the NSM is at the start of the isolating run sequence, 
+    it will get the type of sos. 
+    (Note that in an isolating run sequence, an isolate initiator followed by 
+    an NSM or any type other than PDI must be an overflow isolate initiator.)
   */
   if (types[0] == NSM)
     types[0] = paragraphLevel;
 
   for (i = 1; i < count; i++) {
     if (types[i] == NSM)
-      types[i] = types[i - 1];
-   /* Is this a safe assumption?
-    * I assumed the previous, IS a character.
-    */
+      switch (types[i - 1]) {
+        when LRI or RLI or FSI or PDI:
+          types[i] = ON;
+        otherwise:
+          types[i] = types[i - 1];
+      }
   }
 
  /* Rule (W2)
   * W2. Search backwards from each instance of a European number until the
-  * first strong type (R, L, AL, or sor) is found.  If an AL is found,
+  * first strong type (R, L, AL, or sos) is found.  If an AL is found,
   * change the type of the European number to Arabic number.
   */
   for (i = 0; i < count; i++) {
@@ -1053,22 +847,30 @@ do_bidi(bidi_char * line, int count)
     }
   }
 
- /* Rule (N1)
-  * N1. A sequence of neutrals takes the direction of the surrounding
-  * strong text if the text on both sides has the same direction. European
-  * and Arabic numbers are treated as though they were R.
+ /* Rule (N0) NOT IMPLEMENTED
+  * N0. Handle bracket pairs in isolating run sequences...
   */
-  if (count >= 2 && types[0] == ON) {
+
+ /* Rule (N1)
+  * N1. A sequence of NIs takes the direction of the surrounding
+  * strong text if the text on both sides has the same direction.
+  * European and Arabic numbers are treated as though they were R.
+  // TODO: check
+    European and Arabic numbers act as if they were R in terms of their 
+    influence on NIs. The start-of-sequence (sos) and end-of-sequence (eos) 
+    types are used at isolating run sequence boundaries.
+  */
+  if (count >= 2 && is_NI(types[0])) {
     if ((types[1] == R) || (types[1] == EN) || (types[1] == AN))
       types[0] = R;
     else if (types[1] == L)
       types[0] = L;
   }
   for (i = 1; i < (count - 1); i++) {
-    if (types[i] == ON) {
+    if (is_NI(types[i])) {
       if (types[i - 1] == L) {
         j = i;
-        while (j < (count - 1) && types[j] == ON) {
+        while (j < (count - 1) && is_NI(types[j])) {
           j++;
         }
         if (types[j] == L) {
@@ -1082,7 +884,7 @@ do_bidi(bidi_char * line, int count)
       else if ((types[i - 1] == R) || (types[i - 1] == EN) ||
                (types[i - 1] == AN)) {
         j = i;
-        while (j < (count - 1) && types[j] == ON) {
+        while (j < (count - 1) && is_NI(types[j])) {
           j++;
         }
         if ((types[j] == R) || (types[j] == EN) || (types[j] == AN)) {
@@ -1094,7 +896,7 @@ do_bidi(bidi_char * line, int count)
       }
     }
   }
-  if (count >= 2 && types[count - 1] == ON) {
+  if (count >= 2 && is_NI(types[count - 1])) {
     if (types[count - 2] == R || types[count - 2] == EN ||
         types[count - 2] == AN)
       types[count - 1] = R;
@@ -1103,10 +905,10 @@ do_bidi(bidi_char * line, int count)
   }
 
  /* Rule (N2)
-  * N2. Any remaining neutrals take the embedding direction.
+  * N2. Any remaining NIs take the embedding direction.
   */
   for (i = 0; i < count; i++) {
-    if (types[i] == ON) {
+    if (is_NI(types[i])) {
       if ((levels[i] % 2) == 0)
         types[i] = L;
       else
@@ -1114,6 +916,7 @@ do_bidi(bidi_char * line, int count)
     }
   }
 
+  trace_bidi("<I1");
  /* Rule (I1)
   * I1. For all characters with an even (left-to-right) embedding
   * direction, those of type R go up one level and those of type AN or
@@ -1138,20 +941,24 @@ do_bidi(bidi_char * line, int count)
         levels[i] += 1;
     }
   }
+  trace_bidi(">I2");
 
  /* Rule (L1)
   * L1. On each line, reset the embedding level of the following characters
   * to the paragraph embedding level:
-  *          (1)segment separators, (2)paragraph separators,
-  *           (3)any sequence of whitespace characters preceding
-  *           a segment separator or paragraph separator,
-  *           (4)and any sequence of white space characters
-  *           at the end of the line.
+  *   (1) segment separators,
+  *   (2) paragraph separators,
+  *   (3) any sequence of whitespace characters or isolate markers
+  *       preceding a segment separator or paragraph separator,
+  *   (4) and any sequence of whitespace characters or isolate markers
+  *       at the end of the line.
   * The types of characters used here are the original types, not those
   * modified by the previous phase.
+    N/A: Because a paragraph separator breaks lines, there will be at most 
+    one per line, at the end of that line.
   */
   j = count - 1;
-  while (j > 0 && (getType(line[j].wc) == WS)) {
+  while (j > 0 && (bidi_class_of(j) == WS)) {
     j--;
   }
   if (j < (count - 1)) {
@@ -1159,13 +966,13 @@ do_bidi(bidi_char * line, int count)
       levels[j] = paragraphLevel;
   }
   for (i = 0; i < count; i++) {
-    tempType = getType(line[i].wc);
+    tempType = bidi_class_of(i);
     if (tempType == WS) {
       j = i;
-      while (j < count && (getType(line[j].wc) == WS)) {
+      while (j < count && (bidi_class_of(j) == WS)) {
         j++;
       }
-      if (j == count || getType(line[j].wc) == B || getType(line[j].wc) == S) {
+      if (j == count || bidi_class_of(j) == B || bidi_class_of(j) == S) {
         for (j--; j >= i; j--) {
           levels[j] = paragraphLevel;
         }
@@ -1176,7 +983,7 @@ do_bidi(bidi_char * line, int count)
     }
   }
 
- /* Rule (L4) NOT IMPLEMENTED
+ /* Rule (L4)
   * L4. A character that possesses the mirrored property as specified by
   * Section 4.7, Mirrored, must be depicted by a mirrored glyph if the
   * resolved directionality of that character is R.
@@ -1201,9 +1008,8 @@ do_bidi(bidi_char * line, int count)
     i++;
   }
  /* maximum level in tempType. */
-  while (tempType > 0) {        /* loop from highest level to the least odd, */
-   /* which i assume is 1 */
-    flipThisRun(line, levels, tempType, count);
+  while (tempType > 0) { /* loop from highest level to the least odd */
+    flip_runs(line, levels, tempType, count);
     tempType--;
   }
 
@@ -1214,5 +1020,8 @@ do_bidi(bidi_char * line, int count)
   * process, then the ordering of the marks and the base character must
   * be reversed.
   */
+  // This is not relevant for mintty as the combining characters are kept 
+  // hidden from this algorithm and are maintained transparently to it.
+
   return R;
 }
