@@ -16,6 +16,26 @@ wstrset(wstring *sp, wstring s)
   *sp = memcpy(renewn((wchar *)*sp, size), s, size * sizeof(wchar));
 }
 
+
+char *
+tmpdir()
+{
+  static char * tmpdir = 0;
+  if (!tmpdir || access(tmpdir, W_OK) < 0)
+    tmpdir = getenv("TMP");
+  if (!tmpdir || access(tmpdir, W_OK) < 0)
+    tmpdir = getenv("TEMP");
+  if (!tmpdir || access(tmpdir, W_OK) < 0) {
+    tmpdir = getenv("LOCALAPPDATA");
+    if (tmpdir)
+      tmpdir = asform("%s/Temp", tmpdir);
+  }
+  if (!tmpdir || access(tmpdir, W_OK) < 0)
+    tmpdir = "/usr/tmp";
+  return tmpdir;
+}
+
+
 #if CYGWIN_VERSION_API_MINOR < 70
 
 int
@@ -68,7 +88,7 @@ int iswspace(wint_t wc) { return wc < 0x100 && isspace(wc); }
 
 #if CYGWIN_VERSION_API_MINOR < 91
 
-/* Copyright (C) 2002 by  Red Hat, Incorporated. All rights reserved.
+/* Copyright (C) 2002 by Red Hat, Incorporated. All rights reserved.
  *
  * Permission to use, copy, modify, and distribute this software
  * is freely granted, provided that this notice is preserved.
@@ -122,6 +142,7 @@ argz_stringify(char *argz, size_t argz_len, int sep)
 }
 #endif
 
+
 #if CYGWIN_VERSION_API_MINOR < 93
 
 /*-
@@ -154,7 +175,7 @@ argz_stringify(char *argz, size_t argz_len, int sep)
  */
 
 #include <fcntl.h>
-#include <sys/termios.h>
+#include <termios.h>
 #include <sys/ioctl.h>
 
 #define TTY_NAME_MAX 32
@@ -167,22 +188,22 @@ login_tty(int fd)
 
   if (setsid () == -1)
     return -1;
-  if ((fdname = ttyname (fd)))
+  if ((fdname = ttyname(fd)))
     {
       if (fd != STDIN_FILENO)
-        close (STDIN_FILENO);
+        close(STDIN_FILENO);
       if (fd != STDOUT_FILENO)
-        close (STDOUT_FILENO);
+        close(STDOUT_FILENO);
       if (fd != STDERR_FILENO)
-        close (STDERR_FILENO);
-      newfd = open (fdname, O_RDWR);
-      close (newfd);
+        close(STDERR_FILENO);
+      newfd = open(fdname, O_RDWR);
+      close(newfd);
     }
   dup2 (fd, STDIN_FILENO);
   dup2 (fd, STDOUT_FILENO);
   dup2 (fd, STDERR_FILENO);
   if (fd > 2)
-    close (fd);
+    close(fd);
   return 0;
 }
 
@@ -192,30 +213,32 @@ openpty(int *amaster, int *aslave, char *name,
 {
   int master, slave;
   char pts[TTY_NAME_MAX];
+  int errret = -2;
 
-  if ((master = open ("/dev/ptmx", O_RDWR | O_NOCTTY)) >= 0)
+  if ((master = open("/dev/ptmx", O_RDWR | O_NOCTTY)) >= 0)
     {
-      grantpt (master);
-      unlockpt (master);
-      strcpy (pts, ptsname (master));
-      if ((slave = open (pts, O_RDWR | O_NOCTTY)) >= 0)
+      grantpt(master);
+      unlockpt(master);
+      strcpy(pts, ptsname(master));
+      if ((slave = open(pts, O_RDWR | O_NOCTTY)) >= 0)
         {
           if (amaster)
             *amaster = master;
           if (aslave)
             *aslave = slave;
           if (name)
-            strcpy (name, pts);
+            strcpy(name, pts);
           if (termp)
-            tcsetattr (slave, TCSAFLUSH, termp);
+            tcsetattr(slave, TCSAFLUSH, termp);
           if (winp)
-            ioctl (master, TIOCSWINSZ, (char *) winp);
+            ioctl(master, TIOCSWINSZ, (struct winsize *)winp);
           return 0;
         }
-      close (master);
+      close(master);
+      errret = -3;
     }
   errno = ENOENT;
-  return -1;
+  return errret;
 }
 
 int
@@ -224,20 +247,22 @@ forkpty(int *amaster, char *name,
 {
   int master, slave, pid;
 
-  if (openpty (&master, &slave, name, termp, winp) == -1)
-    return -1;
-  switch (pid = fork ())
+  int ret = openpty(&master, &slave, name, termp, winp);
+  if (ret < 0)
+    return ret;
+
+  switch (pid = fork())
     {
       case -1:
         return -1;
       case 0:
-        close (master);
-        login_tty (slave);
+        close(master);
+        login_tty(slave);
         return 0;
     }
   if (amaster)
     *amaster = master;
-  close (slave);
+  close(slave);
   return pid;
 }
 
