@@ -1,19 +1,26 @@
 // termline.c (part of mintty)
-// Copyright 2008-12 Andy Koppe
+// Copyright 2008-12 Andy Koppe, -2019 Thomas Wolff
 // Adapted from code from PuTTY-0.60 by Simon Tatham and team.
 // Licensed under the terms of the GNU General Public License v3 or later.
 
 #include "termpriv.h"
+#include "win.h"  // cfg.bidi
+
+
+#define newn_1(poi, type, count)	{poi = newn(type, count + 1); poi++;}
+#define renewn_1(poi, count)	{poi--; poi = renewn(poi, count + 1); poi++;}
+
 
 termline *
 newline(int cols, int bce)
 {
   termline *line = new(termline);
-  line->chars = newn(termchar, cols);
-  for (int j = 0; j < cols; j++)
+  newn_1(line->chars, termchar, cols);
+  //! Note: line->chars is based @ index -1
+  for (int j = -1; j < cols; j++)
     line->chars[j] = (bce ? term.erase_char : basic_erase_char);
   line->cols = line->size = cols;
-  line->attr = LATTR_NORM;
+  line->lattr = LATTR_NORM;
   line->temporary = false;
   line->cc_free = 0;
   return line;
@@ -23,7 +30,8 @@ void
 freeline(termline *line)
 {
   assert(line);
-  free(line->chars);
+  //! Note: line->chars is based @ index -1
+  free(&line->chars[-1]);
   free(line);
 }
 
@@ -62,9 +70,9 @@ get(struct buf *b)
  * Add a combining character to a character cell.
  */
 void
-add_cc(termline *line, int col, wchar chr)
+add_cc(termline *line, int col, wchar chr, cattr attr)
 {
-  assert(col >= 0 && col < line->cols);
+  assert(col >= -1 && col < line->cols);
 
  /*
   * Start by extending the cols array if the free list is empty.
@@ -72,7 +80,7 @@ add_cc(termline *line, int col, wchar chr)
   if (!line->cc_free) {
     int n = line->size;
     line->size += 16 + (line->size - line->cols) / 2;
-    line->chars = renewn(line->chars, line->size);
+    renewn_1(line->chars, line->size);
     line->cc_free = n;
     do
       line->chars[n].cc_next = 1;
@@ -87,8 +95,8 @@ add_cc(termline *line, int col, wchar chr)
     col += line->chars[col].cc_next;
 
  /*
-  * `col' now points at the last cc currently in this cell; so
-  * we simply add another one.
+  * `col' now points at the last cc currently in this cell; 
+  * so we simply add another one.
   */
   int newcc = line->cc_free;
   if (line->chars[newcc].cc_next)
@@ -97,6 +105,7 @@ add_cc(termline *line, int col, wchar chr)
     line->cc_free = 0;
   line->chars[newcc].cc_next = 0;
   line->chars[newcc].chr = chr;
+  line->chars[newcc].attr = attr;
   line->chars[col].cc_next = newcc - col;
 }
 
@@ -108,7 +117,7 @@ clear_cc(termline *line, int col)
 {
   int oldfree, origcol = col;
 
-  assert(col >= 0 && col < line->cols);
+  assert(col >= -1 && col < line->cols);
 
   if (!line->chars[col].cc_next)
     return;     /* nothing needs doing */
@@ -126,9 +135,8 @@ clear_cc(termline *line, int col)
 }
 
 /*
- * Compare two character cells for equality. Special case required
- * in do_paint() where we override what we expect the chr and attr
- * fields to be.
+ * Compare two character cells for equality. Special case required in
+ * do_paint() where we override what we expect the chr and attr fields to be.
  */
 int
 termchars_equal_override(termchar *a, termchar *b, uint bchr, cattr battr)
@@ -142,6 +150,8 @@ termchars_equal_override(termchar *a, termchar *b, uint bchr, cattr battr)
     return false;
   if (a->attr.truebg != battr.truebg)
     return false;
+  if (a->attr.ulcolr != battr.ulcolr)
+    return false;
   while (a->cc_next || b->cc_next) {
     if (!a->cc_next || !b->cc_next)
       return false;     /* one cc-list ends, other does not */
@@ -160,8 +170,8 @@ termchars_equal(termchar *a, termchar *b)
 }
 
 /*
- * Copy a character cell. (Requires a pointer to the destination
- * termline, so as to access its free list.)
+ * Copy a character cell. (Requires a pointer to the destination termline,
+ * so as to access its free list.)
  */
 void
 copy_termchar(termline *destline, int x, termchar *src)
@@ -173,7 +183,7 @@ copy_termchar(termline *destline, int x, termchar *src)
 
   while (src->cc_next) {
     src += src->cc_next;
-    add_cc(destline, x, src->chr);
+    add_cc(destline, x, src->chr, src->attr);
   }
 }
 
@@ -239,43 +249,36 @@ makeliteral_attr(struct buf *b, termchar *c)
   * ensures that attribute values remain 16-bit _unless_ the
   * user uses extended colour.
   */
-  uint colourbits;
-  uint attr = c->attr.attr;
+  cattrflags attr = c->attr.attr & ~DATTR_MASK;
   uint truefg = c->attr.truefg;
   uint truebg = c->attr.truebg;
+  colour ulcolr = c->attr.ulcolr;
 
-  assert(ATTR_BGSHIFT > ATTR_FGSHIFT);
-
-  colourbits = (attr >> (ATTR_BGSHIFT + 4)) & 0xF;
-  colourbits <<= 4;
-  colourbits |= (attr >> (ATTR_FGSHIFT + 4)) & 0xF;
-
-  attr =
-    (((attr >> (ATTR_BGSHIFT + 8)) << (ATTR_BGSHIFT + 4)) |
-     (attr & ((1 << (ATTR_BGSHIFT + 4)) - 1)));
-  attr =
-    (((attr >> (ATTR_FGSHIFT + 8)) << (ATTR_FGSHIFT + 4)) |
-     (attr & ((1 << (ATTR_FGSHIFT + 4)) - 1)));
-
-  attr |= (colourbits << (32 - 9));
-
-  if (attr < 0x8000) {
-    add(b, (uchar) ((attr >> 8) & 0xFF));
-    add(b, (uchar) (attr & 0xFF));
-  }
-  else {
-    add(b, (uchar) (((attr >> 24) & 0x7F) | 0x80));
+  if (attr < 0x800000 && !truefg && !truebg && ulcolr != (colour)-1) {
     add(b, (uchar) ((attr >> 16) & 0xFF));
     add(b, (uchar) ((attr >> 8) & 0xFF));
     add(b, (uchar) (attr & 0xFF));
   }
+  else {
+    add(b, (uchar) ((attr >> 56) & 0xFF) | 0x80);
+    add(b, (uchar) ((attr >> 48) & 0xFF));
+    add(b, (uchar) ((attr >> 40) & 0xFF));
+    add(b, (uchar) ((attr >> 32) & 0xFF));
+    add(b, (uchar) ((attr >> 24) & 0xFF));
+    add(b, (uchar) ((attr >> 16) & 0xFF));
+    add(b, (uchar) ((attr >> 8) & 0xFF));
+    add(b, (uchar) (attr & 0xFF));
 
-  add(b, (uchar) ((truefg >> 16) & 0xFF));
-  add(b, (uchar) ((truefg >> 8) & 0xFF));
-  add(b, (uchar) (truefg & 0xFF));
-  add(b, (uchar) ((truebg >> 16) & 0xFF));
-  add(b, (uchar) ((truebg >> 8) & 0xFF));
-  add(b, (uchar) (truebg & 0xFF));
+    add(b, (uchar) ((truefg >> 16) & 0xFF));
+    add(b, (uchar) ((truefg >> 8) & 0xFF));
+    add(b, (uchar) (truefg & 0xFF));
+    add(b, (uchar) ((truebg >> 16) & 0xFF));
+    add(b, (uchar) ((truebg >> 8) & 0xFF));
+    add(b, (uchar) (truebg & 0xFF));
+    add(b, (uchar) ((ulcolr >> 16) & 0xFF));
+    add(b, (uchar) ((ulcolr >> 8) & 0xFF));
+    add(b, (uchar) (ulcolr & 0xFF));
+  }
 }
 
 static void
@@ -284,8 +287,7 @@ makeliteral_cc(struct buf *b, termchar *c)
  /*
   * For combining characters, I just encode a bunch of ordinary
   * chars using makeliteral_chr, and terminate with a \0
-  * character (which I know won't come up as a combining char
-  * itself).
+  * character (which I know won't come up as a combining char itself).
   */
   termchar z;
 
@@ -293,6 +295,7 @@ makeliteral_cc(struct buf *b, termchar *c)
     c += c->cc_next;
     assert(c->chr != 0);
     makeliteral_chr(b, c);
+    makeliteral_attr(b, c);
   }
 
   z.chr = 0;
@@ -321,41 +324,42 @@ readliteral_chr(struct buf *buf, termchar *c, termline *unused(line))
 static void
 readliteral_attr(struct buf *b, termchar *c, termline *unused(line))
 {
-  uint val, attr, colourbits, fg, bg;
+  cattrflags attr;
+  uint fg = 0;
+  uint bg = 0;
+  colour ul = (colour)-1;
 
-  val = get(b) << 8;
-  val |= get(b);
+  attr = get(b) << 16;
+  attr |= get(b) << 8;
+  attr |= get(b);
+  if (attr >= 0x800000) {
+    attr &= ~0x800000;
+    attr <<= 8;
+    attr |= get(b);
+    attr <<= 8;
+    attr |= get(b);
+    attr <<= 8;
+    attr |= get(b);
+    attr <<= 8;
+    attr |= get(b);
+    attr <<= 8;
+    attr |= get(b);
 
-  if (val >= 0x8000) {
-    val &= ~0x8000;
-    val <<= 16;
-    val |= get(b) << 8;
-    val |= get(b);
+    fg = get(b) << 16;
+    fg |= get(b) << 8;
+    fg |= get(b);
+    bg = get(b) << 16;
+    bg |= get(b) << 8;
+    bg |= get(b);
+    ul = get(b) << 16;
+    ul |= get(b) << 8;
+    ul |= get(b);
   }
 
-  colourbits = (val >> (32 - 9)) & 0xFF;
-  attr = (val & ((1 << (32 - 9)) - 1));
-
-  attr =
-    (((attr >> (ATTR_FGSHIFT + 4)) << (ATTR_FGSHIFT + 8)) |
-     (attr & ((1 << (ATTR_FGSHIFT + 4)) - 1)));
-  attr =
-    (((attr >> (ATTR_BGSHIFT + 4)) << (ATTR_BGSHIFT + 8)) |
-     (attr & ((1 << (ATTR_BGSHIFT + 4)) - 1)));
-
-  attr |= (colourbits >> 4) << (ATTR_BGSHIFT + 4);
-  attr |= (colourbits & 0xF) << (ATTR_FGSHIFT + 4);
-
-  fg = get(b) << 16;
-  fg |= get(b) << 8;
-  fg |= get(b);
-  bg = get(b) << 16;
-  bg |= get(b) << 8;
-  bg |= get(b);
-
   c->attr.attr = attr;
   c->attr.truefg = fg;
   c->attr.truebg = bg;
+  c->attr.ulcolr = ul;
 }
 
 static void
@@ -370,7 +374,8 @@ readliteral_cc(struct buf *b, termchar *c, termline *line)
     readliteral_chr(b, &n, line);
     if (!n.chr)
       break;
-    add_cc(line, x, n.chr);
+    readliteral_attr(b, &n, line);
+    add_cc(line, x, n.chr, n.attr);
   }
 }
 
@@ -378,10 +383,14 @@ static void
 makerle(struct buf *b, termline *line,
         void (*makeliteral) (struct buf *b, termchar *c))
 {
-  int hdrpos, hdrsize, n, prevlen, prevpos, thislen, thispos, prev2;
-  termchar *c = line->chars;
+  int hdrpos, hdrsize, prevlen, prevpos, thislen, thispos, prev2;
 
-  n = line->cols;
+  termchar *c = line->chars;
+  int n = line->cols;
+
+  //! Note: line->chars is based @ index -1
+  c--;
+  n++;
 
   hdrpos = b->len;
   hdrsize = 0;
@@ -399,12 +408,11 @@ makerle(struct buf *b, termline *line,
       * This literal precisely matches the previous one.
       * Turn it into a run if it's worthwhile.
       *
-      * With one-byte literals, it costs us two bytes to
-      * encode a run, plus another byte to write the header
-      * to resume normal output; so a three-element run is
-      * neutral, and anything beyond that is unconditionally
-      * worthwhile. With two-byte literals or more, even a
-      * 2-run is a win.
+      * With one-byte literals, it costs us two bytes to encode a run, 
+      * plus another byte to write the header to resume normal output; 
+      * so a three-element run is neutral, and anything beyond that 
+      * is unconditionally worthwhile. 
+      * With two-byte literals or more, even a 2-run is a win.
       */
       if (thislen > 1 || prev2) {
         int runpos, runlen;
@@ -432,8 +440,7 @@ makerle(struct buf *b, termline *line,
           runpos = prevpos;
           b->len = prevpos + prevlen + 1;
          /*
-          * Terminate the previous run of ordinary
-          * literals.
+          * Terminate the previous run of ordinary literals.
           */
           assert(hdrsize >= 1 && hdrsize <= 128);
           b->data[hdrpos] = hdrsize - 1;
@@ -468,9 +475,8 @@ makerle(struct buf *b, termline *line,
       }
       else {
        /*
-        * Just flag that the previous two literals were
-        * identical, in case we find a third identical one
-        * we want to turn into a run.
+        * Just flag that the previous two literals were identical,
+        * in case we find a third identical one we want to turn into a run.
         */
         prev2 = true;
         prevlen = thislen;
@@ -484,8 +490,7 @@ makerle(struct buf *b, termline *line,
     }
 
    /*
-    * This character isn't (yet) part of a run. Add it to
-    * hdrsize.
+    * This character isn't (yet) part of a run. Add it to hdrsize.
     */
     hdrsize++;
     if (hdrsize == 128) {
@@ -515,9 +520,8 @@ compressline(termline *line)
   struct buf buffer = { null, 0, 0 }, *b = &buffer;
 
  /*
-  * First, store the column count, 7 bits at a time, least
-  * significant `digit' first, with the high bit set on all but
-  * the last.
+  * First, store the column count, 7 bits at a time, least significant
+  * `digit' first, with the high bit set on all but the last.
   */
   {
     int n = line->cols;
@@ -532,7 +536,19 @@ compressline(termline *line)
   * Next store the line attributes; same principle.
   */
   {
-    int n = line->attr;
+    int n = line->lattr;
+    while (n >= 128) {
+      add(b, (uchar) ((n & 0x7F) | 0x80));
+      n >>= 7;
+    }
+    add(b, (uchar) (n));
+  }
+
+ /*
+  * Store the wrap position if used.
+  */
+  if (line->lattr & LATTR_WRAPPED) {
+    int n = line->wrappos;
     while (n >= 128) {
       add(b, (uchar) ((n & 0x7F) | 0x80));
       n >>= 7;
@@ -560,6 +576,9 @@ compressline(termline *line)
  /*
   * Trim the allocated memory so we don't waste any, and return.
   */
+#ifdef debug_compressline
+  printf("compress %d chars -> %d bytes\n", line->size, b->len);
+#endif
   return renewn(b->data, b->len);
 }
 
@@ -567,7 +586,8 @@ static void
 readrle(struct buf *b, termline *line,
         void (*readliteral) (struct buf *b, termchar *c, termline *line))
 {
-  int n = 0;
+  //! Note: line->chars is based @ index -1
+  int n = -1;
 
   while (n < line->cols) {
     int hdr = get(b);
@@ -622,33 +642,43 @@ decompressline(uchar *data, int *bytes_used)
   * Now create the output termline.
   */
   line = new(termline);
-  line->chars = newn(termchar, ncols);
+  newn_1(line->chars, termchar, ncols);
   line->cols = line->size = ncols;
   line->temporary = true;
   line->cc_free = 0;
 
  /*
-  * We must set all the cc pointers in line->chars to 0 right
-  * now, so that cc diagnostics that verify the integrity of the
-  * whole line will make sense while we're in the middle of
-  * building it up.
+  * We must set all the cc pointers in line->chars to 0 right now, 
+  * so that cc diagnostics that verify the integrity of the whole line 
+  * will make sense while we're in the middle of building it up.
   */
-  {
-    int i;
-    for (i = 0; i < line->cols; i++)
-      line->chars[i].cc_next = 0;
-  }
+  //! Note: line->chars is based @ index -1
+  for (int i = -1; i < line->cols; i++)
+    line->chars[i].cc_next = 0;
 
  /*
   * Now read in the line attributes.
   */
-  line->attr = shift = 0;
+  line->lattr = shift = 0;
   do {
     byte = get(b);
-    line->attr |= (byte & 0x7F) << shift;
+    line->lattr |= (byte & 0x7F) << shift;
     shift += 7;
   } while (byte & 0x80);
 
+ /*
+  * Read the wrap position if used.
+  */
+  if (line->lattr & LATTR_WRAPPED) {
+    ncols = shift = 0;
+    do {
+      byte = get(b);
+      ncols |= (byte & 0x7F) << shift;
+      shift += 7;
+    } while (byte & 0x80);
+    line->wrappos = ncols;
+  }
+
  /*
   * Now we read in each of the RLE streams in turn.
   */
@@ -669,12 +699,13 @@ decompressline(uchar *data, int *bytes_used)
 void
 clearline(termline *line)
 {
-  line->attr = LATTR_NORM;
-  for (int j = 0; j < line->cols; j++)
+  line->lattr = LATTR_NORM;
+  //! Note: line->chars is based @ index -1
+  for (int j = -1; j < line->cols; j++)
     line->chars[j] = term.erase_char;
   if (line->size > line->cols) {
     line->size = line->cols;
-    line->chars = renewn(line->chars, line->size);
+    renewn_1(line->chars, line->size);
     line->cc_free = 0;
   }
 }
@@ -693,7 +724,7 @@ resizeline(termline *line, int cols)
     * Leave the same amount of cc space as there was to begin with.
     */
     line->size += cols - oldcols;
-    line->chars = renewn(line->chars, line->size);
+    renewn_1(line->chars, line->size);
     line->cols = cols;
 
    /*
@@ -708,7 +739,8 @@ resizeline(termline *line, int cols)
     * relative offsets within the cc block.) Also do the same
     * to the head of the cc_free list.
     */
-    for (int i = 0; i < oldcols; i++)
+    //! Note: line->chars is based @ index -1
+    for (int i = -1; i < oldcols; i++)
       if (line->chars[i].cc_next)
         line->chars[i].cc_next += cols - oldcols;
     if (line->cc_free)
@@ -735,8 +767,7 @@ sblines(void)
 
 /*
  * Retrieve a line of the screen or of the scrollback, according to
- * whether the y coordinate is non-negative or negative
- * (respectively).
+ * whether the y coordinate is non-negative or negative (respectively).
  */
 termline *
 fetch_line(int y)
@@ -798,12 +829,12 @@ term_bidi_cache_hit(int line, termchar *lbefore, int width)
     if (!termchars_equal(term.pre_bidi_cache[line].chars + i, lbefore + i))
       return false;     /* line doesn't match cache */
 
-  return true;  /* it didn't match. */
+  return true;  /* all termchars matched */
 }
 
 static void
 term_bidi_cache_store(int line, termchar *lbefore, termchar *lafter,
-                      bidi_char *wcTo, int width, int size)
+                      bidi_char *wcTo, int width, int size, int bidisize)
 {
   int i;
 
@@ -835,19 +866,39 @@ term_bidi_cache_store(int line, termchar *lbefore, termchar *lafter,
 
   memcpy(term.pre_bidi_cache[line].chars, lbefore, size * sizeof(termchar));
   memcpy(term.post_bidi_cache[line].chars, lafter, size * sizeof(termchar));
-  memset(term.post_bidi_cache[line].forward, 0, width * sizeof (int));
-  memset(term.post_bidi_cache[line].backward, 0, width * sizeof (int));
+  memset(term.post_bidi_cache[line].forward, 0, width * sizeof(int));
+  memset(term.post_bidi_cache[line].backward, 0, width * sizeof(int));
 
+  int ib = 0;
   for (i = 0; i < width; i++) {
-    int p = wcTo[i].index;
+    while (wcTo[ib].index == -1)
+      ib++;
+
+    int p = wcTo[ib].index;
 
     assert(0 <= p && p < width);
 
     term.post_bidi_cache[line].backward[i] = p;
     term.post_bidi_cache[line].forward[p] = i;
+
+    ib++;
   }
+  (void)bidisize;
+  assert(ib == bidisize);
 }
 
+#ifdef debug_bidi
+void trace_bidi(char * tag, bidi_char * wc)
+{
+  printf("%s", tag);
+  for (int i = 0; i < term.cols; i++)
+    if (wc[i].wc != ' ') printf(" %04X", wc[i].wc);
+  printf("\n");
+}
+#else
+#define trace_bidi(tag, wc)	
+#endif
+
 /*
  * Prepare the bidi information for a screen line. Returns the
  * transformed list of termchars, or null if no transformation at
@@ -859,8 +910,14 @@ term_bidi_cache_store(int line, termchar *lbefore, termchar *lafter,
 termchar *
 term_bidi_line(termline *line, int scr_y)
 {
+  if ((line->lattr & LATTR_NOBIDI) || term.disable_bidi
+      || cfg.bidi == 0
+      || (cfg.bidi == 1 && (term.on_alt_screen ^ term.show_other_screen))
+     )
+    return null;
+
   termchar *lchars;
-  int it;
+  int it, ib;
 
  /* Do Arabic shaping and bidi. */
 
@@ -872,14 +929,83 @@ term_bidi_line(termline *line, int scr_y)
       term.wcTo = renewn(term.wcTo, term.wcFromTo_size);
     }
 
+    // UTF-16 string (including surrogates) for Windows bidi calculation
+    //wchar wcs[2 * term.cols];  /// size handling to be tweaked
+    //int wcsi = 0;
+
+    ib = 0;
     for (it = 0; it < term.cols; it++) {
-      wchar c = line->chars[it].chr;
-      term.wcFrom[it].origwc = term.wcFrom[it].wc = c;
-      term.wcFrom[it].index = it;
+      ucschar c = line->chars[it].chr;
+      //wcs[wcsi++] = c;
+
+      if ((c & 0xFC00) == 0xD800) {
+        int off = line->chars[it].cc_next;
+        if (off) {
+          ucschar low_surrogate = line->chars[it + off].chr;
+          if ((low_surrogate & 0xFC00) == 0xDC00) {
+            c = ((c - 0xD7C0) << 10) | (low_surrogate & 0x03FF);
+            //wcs[wcsi++] = low_surrogate;
+          }
+        }
+      }
+
+      if (!it) {
+        // Handle initial combining characters, esp. directional markers
+        termchar * bp = &line->chars[-1];
+        // Unfold directional formatting characters which are handled 
+        // like combining characters in the mintty structures 
+        // (and would thus stay hidden from minibidi), and need to be 
+        // exposed as separate characters for the minibidi algorithm
+        while (bp->cc_next) {
+          bp += bp->cc_next;
+          if (bp->chr == 0x200E || bp->chr == 0x200F
+              || (bp->chr >= 0x202A && bp->chr <= 0x202E)
+              || (bp->chr >= 0x2066 && bp->chr <= 0x2069)
+             )
+          {
+            term.wcFromTo_size++;
+            term.wcFrom = renewn(term.wcFrom, term.wcFromTo_size);
+            term.wcTo = renewn(term.wcTo, term.wcFromTo_size);
+            term.wcFrom[ib].origwc = term.wcFrom[ib].wc = bp->chr;
+            term.wcFrom[ib].index = -1;
+            ib++;
+            //wcs[wcsi++] = bp->chr;
+          }
+        }
+      }
+
+      term.wcFrom[ib].origwc = term.wcFrom[ib].wc = c;
+      term.wcFrom[ib].index = it;
+      ib++;
+
+      termchar * bp = &line->chars[it];
+      // Unfold directional formatting characters which are handled 
+      // like combining characters in the mintty structures 
+      // (and would thus stay hidden from minibidi), and need to be 
+      // exposed as separate characters for the minibidi algorithm
+      while (bp->cc_next) {
+        bp += bp->cc_next;
+        if (bp->chr == 0x200E || bp->chr == 0x200F
+            || (bp->chr >= 0x202A && bp->chr <= 0x202E)
+            || (bp->chr >= 0x2066 && bp->chr <= 0x2069)
+           )
+        {
+          term.wcFromTo_size++;
+          term.wcFrom = renewn(term.wcFrom, term.wcFromTo_size);
+          term.wcTo = renewn(term.wcTo, term.wcFromTo_size);
+          term.wcFrom[ib].origwc = term.wcFrom[ib].wc = bp->chr;
+          term.wcFrom[ib].index = -1;
+          ib++;
+          //wcs[wcsi++] = bp->chr;
+        }
+      }
     }
 
-    do_bidi(term.wcFrom, term.cols);
-    do_shape(term.wcFrom, term.wcTo, term.cols);
+    trace_bidi("=", term.wcFrom);
+    do_bidi(term.wcFrom, ib);
+    trace_bidi(":", term.wcFrom);
+    do_shape(term.wcFrom, term.wcTo, ib);
+    trace_bidi("~", term.wcTo);
 
     if (term.ltemp_size < line->size) {
       term.ltemp_size = line->size;
@@ -888,16 +1014,22 @@ term_bidi_line(termline *line, int scr_y)
 
     memcpy(term.ltemp, line->chars, line->size * sizeof(termchar));
 
+    ib = 0;
     for (it = 0; it < term.cols; it++) {
-      term.ltemp[it] = line->chars[term.wcTo[it].index];
-      if (term.ltemp[it].cc_next)
-        term.ltemp[it].cc_next -= it - term.wcTo[it].index;
+      while (term.wcTo[ib].index == -1)
+        ib++;
 
-      if (term.wcTo[it].origwc != term.wcTo[it].wc)
-        term.ltemp[it].chr = term.wcTo[it].wc;
+      term.ltemp[it] = line->chars[term.wcTo[ib].index];
+      if (term.ltemp[it].cc_next)
+        term.ltemp[it].cc_next -= it - term.wcTo[ib].index;
+
+      if (term.wcTo[ib].origwc != term.wcTo[ib].wc)
+        term.ltemp[it].chr = term.wcTo[ib].wc;
+
+      ib++;
     }
     term_bidi_cache_store(scr_y, line->chars, term.ltemp, term.wcTo,
-                          term.cols, line->size);
+                          term.cols, line->size, ib);
 
     lchars = term.ltemp;
   }
