@@ -1,5 +1,5 @@
 // termmouse.c (part of mintty)
-// Copyright 2008-12 Andy Koppe
+// Copyright 2008-12 Andy Koppe, 2017 Thomas Wolff
 // Based on code from PuTTY-0.60 by Simon Tatham and team.
 // Licensed under the terms of the GNU General Public License v3 or later.
 
@@ -46,13 +46,15 @@ sel_spread_word(pos p, bool forward)
       if (!forward)
         ret_p = p;
     }
+    else if (c == ' ' && p.x > 0 && get_char(line, p.x - 1) == '\\')
+      ret_p = p;
     else if (!(strchr("&,;?!", c) || c == (forward ? '=' : ':')))
       break;
 
     if (forward) {
       p.x++;
-      if (p.x >= term.cols - ((line->attr & LATTR_WRAPPED2) != 0)) {
-        if (!(line->attr & LATTR_WRAPPED))
+      if (p.x >= term.cols - ((line->lattr & LATTR_WRAPPED2) != 0)) {
+        if (!(line->lattr & LATTR_WRAPPED))
           break;
         p.x = 0;
         release_line(line);
@@ -65,9 +67,9 @@ sel_spread_word(pos p, bool forward)
           break;
         release_line(line);
         line = fetch_line(--p.y);
-        if (!(line->attr & LATTR_WRAPPED))
+        if (!(line->lattr & LATTR_WRAPPED))
           break;
-        p.x = term.cols - ((line->attr & LATTR_WRAPPED2) != 0);
+        p.x = term.cols - ((line->lattr & LATTR_WRAPPED2) != 0);
       }
       p.x--;
     }
@@ -90,7 +92,7 @@ sel_spread_half(pos p, bool forward)
       * for runs of spaces at the end of a non-wrapping line.
       */
       termline *line = fetch_line(p.y);
-      if (!(line->attr & LATTR_WRAPPED)) {
+      if (!(line->lattr & LATTR_WRAPPED)) {
         termchar *q = line->chars + term.cols;
         while (q > line->chars && q[-1].chr == ' ' && !q[-1].cc_next)
           q--;
@@ -106,7 +108,7 @@ sel_spread_half(pos p, bool forward)
     when MS_SEL_LINE:
       if (forward) {
         termline *line = fetch_line(p.y);
-        while (line->attr & LATTR_WRAPPED) {
+        while (line->lattr & LATTR_WRAPPED) {
           release_line(line);
           line = fetch_line(++p.y);
           p.x = 0;
@@ -123,14 +125,14 @@ sel_spread_half(pos p, bool forward)
         p.x = 0;
         while (p.y > -sblines()) {
           termline *line = fetch_line(p.y - 1);
-          bool wrapped = line->attr & LATTR_WRAPPED;
+          bool wrapped = line->lattr & LATTR_WRAPPED;
           release_line(line);
           if (!wrapped)
             break;
           p.y--;
         }
       }
-    default:
+    otherwise:
      /* Shouldn't happen. */
       break;
   }
@@ -145,9 +147,20 @@ sel_spread(void)
   incpos(term.sel_end);
 }
 
+static bool
+hover_spread_empty(void)
+{
+  term.hover_start = sel_spread_word(term.hover_start, false);
+  term.hover_end = sel_spread_word(term.hover_end, true);
+  bool eq = term.hover_start.y == term.hover_end.y && term.hover_start.x == term.hover_end.x;
+  incpos(term.hover_end);
+  return eq;
+}
+
 static void
 sel_drag(pos selpoint)
 {
+  //printf("sel_drag %d+%d/2 (anchor %d+%d/2)\n", selpoint.x, selpoint.r, term.sel_anchor.x, term.sel_anchor.r);
   term.selected = true;
   if (!term.sel_rect) {
    /*
@@ -157,10 +170,26 @@ sel_drag(pos selpoint)
     if (poslt(selpoint, term.sel_anchor)) {
       term.sel_start = selpoint;
       term.sel_end = term.sel_anchor;
+      if (cfg.elastic_mouse && !term.mouse_mode) {
+        if (selpoint.r) {
+          incpos(term.sel_start);
+        }
+        if (!term.sel_anchor.r) {
+          decpos(term.sel_end);
+        }
+      }
     }
     else {
       term.sel_start = term.sel_anchor;
       term.sel_end = selpoint;
+      if (cfg.elastic_mouse && !term.mouse_mode) {
+        if (term.sel_anchor.r) {
+          incpos(term.sel_start);
+        }
+        if (!selpoint.r) {
+          decpos(term.sel_end);
+        }
+      }
     }
     sel_spread();
   }
@@ -180,6 +209,7 @@ sel_drag(pos selpoint)
 static void
 sel_extend(pos selpoint)
 {
+  //printf("sel_extend %d+%d/2 (anchor %d+%d/2)\n", selpoint.x, selpoint.r, term.sel_anchor.x, term.sel_anchor.r);
   if (term.selected) {
     if (!term.sel_rect) {
      /*
@@ -221,13 +251,61 @@ typedef enum {
   MA_MOVE = 1,
   MA_WHEEL = 2,
   MA_RELEASE = 3
-} mouse_action;
+} mouse_action;  // values are significant, used for calculation!
 
 static void
 send_mouse_event(mouse_action a, mouse_button b, mod_keys mods, pos p)
 {
+  if (term.mouse_mode == MM_LOCATOR) {
+    // handle DECSLE: select locator events
+    if ((a == MA_CLICK && term.locator_report_up)
+     || (a == MA_RELEASE && term.locator_report_dn)) {
+      int pe = 0;
+      switch (b) {
+        when MBT_LEFT:
+          pe = a == MA_CLICK ? 2 : 3;
+        when MBT_MIDDLE:
+          pe = a == MA_CLICK ? 4 : 5;
+        when MBT_RIGHT:
+          pe = a == MA_CLICK ? 6 : 7;
+        when MBT_4:
+          pe = a == MA_CLICK ? 8 : 9;
+        otherwise:;
+      }
+      if (pe) {
+        int x, y, buttons;
+        win_get_locator_info(&x, &y, &buttons, term.locator_by_pixels);
+        child_printf("\e[%d;%d;%d;%d;0&w", pe, buttons, y, x);
+        term.locator_rectangle = false;
+      }
+    }
+    // handle DECEFR: enable filter rectangle
+    else if (a == MA_MOVE && term.locator_rectangle) {
+      /* Anytime the locator is detected outside of the filter
+         rectangle, an outside rectangle event is generated and the
+         rectangle is disabled.
+      */
+      int x, y, buttons;
+      win_get_locator_info(&x, &y, &buttons, term.locator_by_pixels);
+      if (x < term.locator_left || x > term.locator_right
+          || y < term.locator_top || y > term.locator_bottom) {
+        child_printf("\e[10;%d;%d;%d;0&w", buttons, y, x);
+        term.locator_rectangle = false;
+      }
+    }
+    return;
+  }
+
   uint x = p.x + 1, y = p.y + 1;
 
+  switch (b) {
+    when MBT_4:
+      b = MBT_LEFT; mods |= MDK_ALT;
+    when MBT_5:
+      b = MBT_RIGHT; mods |= MDK_ALT;
+    otherwise:;
+  }
+
   uint code = b ? b - 1 : 0x3;
 
   if (a != MA_RELEASE)
@@ -283,9 +361,9 @@ box_pos(pos p)
 static pos
 get_selpoint(const pos p)
 {
-  pos sp = { .y = p.y + term.disptop, .x = p.x };
+  pos sp = { .y = p.y + term.disptop, .x = p.x, .r = p.r };
   termline *line = fetch_line(sp.y);
-  if ((line->attr & LATTR_MODE) != LATTR_NORM)
+  if ((line->lattr & LATTR_MODE) != LATTR_NORM)
     sp.x /= 2;
 
  /*
@@ -318,6 +396,8 @@ send_keys(char *code, uint len, uint count)
 static bool
 is_app_mouse(mod_keys *mods_p)
 {
+  if (term.locator_1_enabled)
+    return true;
   if (!term.mouse_mode || term.show_other_screen)
     return false;
   bool override = *mods_p & cfg.click_target_mod;
@@ -328,6 +408,11 @@ is_app_mouse(mod_keys *mods_p)
 void
 term_mouse_click(mouse_button b, mod_keys mods, pos p, int count)
 {
+  if (term.hovering) {
+    term.hovering = false;
+    win_update(true);
+  }
+
   if (is_app_mouse(&mods)) {
     if (term.mouse_mode == MM_X10)
       mods = 0;
@@ -335,14 +420,29 @@ term_mouse_click(mouse_button b, mod_keys mods, pos p, int count)
     term.mouse_state = b;
   }
   else {
+    // generic transformation M4/M5 -> Alt+left/right;
+    // if any specific handling is designed for M4/M5, this needs to be tweaked
+    bool fake_alt = false;
+    switch (b) {
+      when MBT_4:
+        b = MBT_LEFT; mods |= MDK_ALT; fake_alt = true;
+      when MBT_5:
+        b = MBT_RIGHT; mods |= MDK_ALT; fake_alt = true;
+      otherwise:;
+    }
+
     bool alt = mods & MDK_ALT;
     bool shift_or_ctrl = mods & (MDK_SHIFT | MDK_CTRL);
     int mca = cfg.middle_click_action;
     int rca = cfg.right_click_action;
     term.mouse_state = 0;
     if (b == MBT_RIGHT && (rca == RC_MENU || shift_or_ctrl)) {
-      if (!alt)
-        win_popup_menu();
+      // disable Alt+mouse menu opening;
+      // the menu would often be closed soon by auto-repeat Alt, sending
+      // WM_CAPTURECHANGED, WM_UNINITMENUPOPUP, WM_MENUSELECT, WM_EXITMENULOOP
+      // trying to ignore WM_CAPTURECHANGED does not help
+      if (!alt || fake_alt)
+        win_popup_menu(mods);
     }
     else if (b == MBT_MIDDLE && (mods & ~MDK_SHIFT) == MDK_CTRL) {
       if (cfg.zoom_mouse)
@@ -360,17 +460,19 @@ term_mouse_click(mouse_button b, mod_keys mods, pos p, int count)
     else if (b == MBT_LEFT && mods == MDK_SHIFT && rca == RC_EXTEND)
       term.mouse_state = MS_PASTING;
     else if (b == MBT_LEFT && mods == MDK_CTRL) {
-      // Open word under cursor
-      p = get_selpoint(box_pos(p));
-      term.mouse_state = MS_OPENING;
-      term.selected = true;
-      term.sel_rect = false;
-      term.sel_start = term.sel_end = term.sel_anchor = p;
-      sel_spread();
-      win_update();
+      if (count == cfg.opening_clicks) {
+        // Open word under cursor
+        p = get_selpoint(box_pos(p));
+        term.mouse_state = MS_OPENING;
+        term.selected = true;
+        term.sel_rect = false;
+        term.sel_start = term.sel_end = term.sel_anchor = p;
+        sel_spread();
+        win_update(true);
+      }
+    }
+    else if (b == MBT_MIDDLE && mca == MC_VOID) {
     }
-    else if (b == MBT_MIDDLE && mca == MC_VOID)
-      (void) 0;
     else {
       // Only clicks for selecting and extending should get here.
       p = get_selpoint(box_pos(p));
@@ -390,7 +492,7 @@ term_mouse_click(mouse_button b, mod_keys mods, pos p, int count)
         sel_spread();
       }
       win_capture_mouse();
-      win_update();
+      win_update(true);
     }
   }
 }
@@ -406,7 +508,8 @@ term_mouse_release(mouse_button b, mod_keys mods, pos p)
     when MS_OPENING:
       term_open();
       term.selected = false;
-      win_update();
+      term.hovering = false;
+      win_update(true);
     when MS_SEL_CHAR or MS_SEL_WORD or MS_SEL_LINE: {
       // Finish selection.
       if (term.selected && cfg.copy_on_select)
@@ -439,12 +542,12 @@ term_mouse_release(mouse_button b, mod_keys mods, pos p)
       uint count = 0;
       while (p.y != end.y) {
         termline *line = fetch_line(p.y);
-        if (!(line->attr & LATTR_WRAPPED)) {
+        if (!(line->lattr & LATTR_WRAPPED)) {
           release_line(line);
           moved_previously = false;
           return;
         }
-        int cols = term.cols - ((line->attr & LATTR_WRAPPED2) != 0);
+        int cols = term.cols - ((line->lattr & LATTR_WRAPPED2) != 0);
         for (int x = p.x; x < cols; x++) {
           if (line->chars[x].chr != UCSWIDE)
             count++;
@@ -468,7 +571,7 @@ term_mouse_release(mouse_button b, mod_keys mods, pos p)
       moved_previously = true;
       last_dest = dest;
     }
-    default:
+    otherwise:
       if (is_app_mouse(&mods)) {
         if (term.mouse_mode >= MM_VT200)
           send_mouse_event(MA_RELEASE, b, mods, box_pos(p));
@@ -482,7 +585,7 @@ sel_scroll_cb(void)
   if (term_selecting() && term.sel_scroll) {
     term_scroll(0, term.sel_scroll);
     sel_drag(get_selpoint(term.sel_pos));
-    win_update();
+    win_update(true);
     win_set_timer(sel_scroll_cb, 125);
   }
 }
@@ -490,7 +593,9 @@ sel_scroll_cb(void)
 void
 term_mouse_move(mod_keys mods, pos p)
 {
+  //printf("mouse_move %d+%d/2\n", p.x, p.r);
   pos bp = box_pos(p);
+
   if (term_selecting()) {
     if (p.y < 0 || p.y >= term.rows) {
       if (!term.sel_scroll)
@@ -501,15 +606,19 @@ term_mouse_move(mod_keys mods, pos p)
     else {
       term.sel_scroll = 0;
       if (p.x < 0 && p.y + term.disptop > term.sel_anchor.y)
-        bp = (pos){.y = p.y - 1, .x = term.cols - 1};
+        bp = (pos){.y = p.y - 1, .x = term.cols - 1, .r = p.r};
     }
+
+    bool alt = mods & MDK_ALT;
+    term.sel_rect = alt;
     sel_drag(get_selpoint(bp));
-    win_update();
+
+    win_update(true);
   }
   else if (term.mouse_state == MS_OPENING) {
     term.mouse_state = 0;
     term.selected = false;
-    win_update();
+    win_update(true);
   }
   else if (term.mouse_state > 0) {
     if (term.mouse_mode >= MM_BTN_EVENT)
@@ -519,17 +628,37 @@ term_mouse_move(mod_keys mods, pos p)
     if (term.mouse_mode == MM_ANY_EVENT)
       send_mouse_event(MA_MOVE, 0, mods, bp);
   }
+
+  if (mods == MDK_CTRL && term.has_focus) {
+    p = get_selpoint(box_pos(p));
+    term.hover_start = term.hover_end = p;
+    if (!hover_spread_empty()) {
+      term.hovering = true;
+      win_update(true);
+    }
+    else if (term.hovering) {
+      term.hovering = false;
+      win_update(true);
+    }
+  }
 }
 
 void
 term_mouse_wheel(int delta, int lines_per_notch, mod_keys mods, pos p)
 {
+  if (term.hovering) {
+    term.hovering = false;
+    win_update(true);
+  }
+
   enum { NOTCH_DELTA = 120 };
 
   static int accu;
   accu += delta;
 
   if (is_app_mouse(&mods)) {
+    if (strstr(cfg.suppress_wheel, "report"))
+      return;
     // Send as mouse events, with one event per notch.
     int notches = accu / NOTCH_DELTA;
     if (notches) {
@@ -540,6 +669,8 @@ term_mouse_wheel(int delta, int lines_per_notch, mod_keys mods, pos p)
     }
   }
   else if ((mods & ~MDK_SHIFT) == MDK_CTRL) {
+    if (strstr(cfg.suppress_wheel, "zoom"))
+      return;
     if (cfg.zoom_mouse) {
       int zoom = accu / NOTCH_DELTA;
       if (zoom) {
@@ -557,9 +688,14 @@ term_mouse_wheel(int delta, int lines_per_notch, mod_keys mods, pos p)
     int lines = lines_per_notch * accu / NOTCH_DELTA;
     if (lines) {
       accu -= lines * NOTCH_DELTA / lines_per_notch;
-      if (!term.on_alt_screen || term.show_other_screen)
+      if (!term.on_alt_screen || term.show_other_screen) {
+        if (strstr(cfg.suppress_wheel, "scrollwin"))
+          return;
         term_scroll(0, -lines);
+      }
       else if (term.wheel_reporting) {
+        if (strstr(cfg.suppress_wheel, "scrollapp"))
+          return;
         // Send scroll distance as CSI a/b events
         bool up = lines > 0;
         lines = abs(lines);
