@@ -1,7 +1,6 @@
 #include "print.h"
 #include "charset.h"
-#include "win.h"  // win_prefix_title, win_unprefix_title
-#include <sys/cygwin.h>
+#include "winpriv.h"  // win_prefix_title, win_unprefix_title
 #include <fcntl.h>
 
 
@@ -14,17 +13,7 @@ static int np = 0;
 void
 printer_start_job(wstring printer_name)
 {
-  char * dirsuf = "";
-  char * tmpdir = getenv("TMP");
-  if (!tmpdir)
-    tmpdir = getenv("TEMP");
-  if (!tmpdir) {
-    tmpdir = getenv("$LOCALAPPDATA");
-    if (tmpdir)
-      dirsuf = "/Temp";
-  }
-  if (!tmpdir)
-    tmpdir = "/usr/tmp";
+  char * tempdir = tmpdir();
 
   char * user = getenv("USER");
   if (!user)
@@ -38,20 +27,34 @@ printer_start_job(wstring printer_name)
   char n[11];
   sprintf(n, "%d", ++np);
 
-  // compose $tmpdir/$dirsuf/mintty.print.$USER.$$
+  // compose $tempdir/mintty.print.$USER.$$
   char * pref = "mintty-print.";
-  pf = malloc(strlen(tmpdir) + strlen(dirsuf) + strlen(pref) + strlen(user) + strlen(pid) + strlen(n) + 7);
-  sprintf(pf, "%s%s/%s%s.%s-%s.prn", tmpdir, dirsuf, pref, user, pid, n);
+  pf = malloc(strlen(tempdir) + strlen(pref) + strlen(user) + strlen(pid) + strlen(n) + 7);
+  sprintf(pf, "%s/%s%s.%s-%s.prn", tempdir, pref, user, pid, n);
 
   pd = open(pf, O_CREAT | O_TRUNC | O_BINARY | O_WRONLY, 0600);
   if (pd >= 0) {
-    win_prefix_title(L"[Printing...] ");
+    win_prefix_title(_W("[Printing...] "));
 
     printer = printer_name;
     write(pd, &BOM, 2);
   }
 }
 
+void
+printer_wwrite(wchar * wdata, uint len)
+{
+  if (printer) {
+    void * wbuf = wdata;
+    uint wlen = len * sizeof(wchar);
+    uint n;
+    while ((n = write(pd, wbuf, wlen)) > 0) {
+      wbuf += n;
+      wlen -= n;
+    }
+  }
+}
+
 void
 printer_write(char * data, uint len)
 {
@@ -89,7 +92,7 @@ printer_finish_job(void)
 
     // retrieve current value of `chcp`
     // which is not necessarily the same as GetOEMCP() !
-    FILE * chcpcom = popen ("chcp.com | sed -e 's,.*:,,' -e 's, ,,'", "r");
+    FILE * chcpcom = popen("chcp.com | sed -e 's,.*:,,' -e 's, ,,'", "r");
     char line[99];
     fgets(line, sizeof line, chcpcom);
     pclose(chcpcom);
@@ -109,6 +112,6 @@ printer_finish_job(void)
     free(pf);
     printer = 0;
 
-    win_unprefix_title(L"[Printing...] ");
+    win_unprefix_title(_W("[Printing...] "));
   }
 }
