@@ -4,14 +4,32 @@
 
 #include "winsearch.h"
 
-void
-scroll_to_result()
+
+static bool search_initialised = false;
+static int prev_height = 0;
+
+static HWND search_wnd;
+static HWND search_close_wnd;
+static HWND search_prev_wnd;
+static HWND search_next_wnd;
+static HWND search_edit_wnd;
+static WNDPROC default_edit_proc;
+static HFONT search_font;
+
+static void win_hide_search(void);
+
+#define SEARCHBARCLASS "SearchBar"
+
+int SEARCHBAR_HEIGHT = 26;
+
+
+static int
+current_delta(void)
 {
   if (term.results.length == 0) {
-    return;
+    return 0;
   }
 
-  // If visible, don't do anything.
   result * res = term.results.results + term.results.current;
   int y = res->y - term.sblines;
   int delta = 0;
@@ -21,44 +39,43 @@ scroll_to_result()
   if (y >= term.disptop + term.rows) {
     delta = y - (term.disptop + term.rows - 1);
   }
+  return delta;
+}
+
+static void
+scroll_to_result(void)
+{
+  if (term.results.length == 0) {
+    return;
+  }
+
+  int delta = current_delta();
+
   // Scroll if we must!
   if (delta != 0) {
     term_scroll(0, delta);
   }
 }
 
-void
+static void
 next_result(void)
 {
   if (term.results.length == 0) {
     return;
   }
-  if (searched) {
+  if (current_delta() == 0)
     term.results.current = (term.results.current + 1) % term.results.length;
-  }
-  else {
-    searched = true;
-  }
   scroll_to_result();
 }
 
-void
+static void
 prev_result(void)
 {
   if (term.results.length == 0) {
     return;
   }
-  if (searched) {
-    term.results.current = (term.results.current - 1) % term.results.length;
-    // Handle wrap arounds correctly.
-    if (term.results.current < 0) {
-      term.results.current += term.results.length;
-    }
-  }
-  else {
-    term.results.current = term.results.length - 1;
-    searched = true;
-  }
+  if (current_delta() == 0)
+    term.results.current = (term.results.current + term.results.length - 1) % term.results.length;
   scroll_to_result();
 }
 
@@ -142,7 +159,6 @@ search_proc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
     GetWindowTextW(search_edit_wnd, buf, len);
     term_set_search(buf);
     term_update_search();
-    searched = false;
     win_schedule_update();
     return 0;
   }
@@ -159,18 +175,24 @@ place_field(int * curpoi, int width, int * pospoi)
   }
 }
 
-void
+static void
 win_toggle_search(bool show, bool focus)
 {
   RECT cr;
   GetClientRect(wnd, &cr);
-
   int width = cr.right - cr.left;
-  int height = SEARCHBAR_HEIGHT;
-  int margin = 2;
-  int button_width = 20;
+
+  int margin = cell_width / 6 + 1;
+  int height = cell_height + margin * 2;
+  int button_width = cell_width * 2;
+  SEARCHBAR_HEIGHT = height;
+
   int edit_width = width - button_width * 3 - margin * 2;
   int ctrl_height = height - margin * 2;
+  int sf_height = ctrl_height - 4;
+#ifdef debug_searchbar
+  printf("ctrl/but %d/%d cell %d/%d font h %d s %d\n", ctrl_height, button_width, cell_height, cell_width, font_height, font_size);
+#endif
 
   const char * search_bar = cfg.search_bar;
   int pos_close = -1;
@@ -195,36 +217,43 @@ win_toggle_search(bool show, bool focus)
   place_field(& barpos, edit_width, & pos_edit);
 
   // Set up our global variables.
-  if (!search_initialised) {
-    RegisterClass(&(WNDCLASS){
-      .style = 0,
-      .lpfnWndProc = search_proc,
-      .cbClsExtra = 0,
-      .cbWndExtra = 0,
-      .hInstance = inst,
-      .hIcon = NULL,
-      .hCursor = NULL,
-      .hbrBackground = (HBRUSH)(COLOR_3DFACE + 1),
-      .lpszMenuName = NULL,
-      .lpszClassName = SEARCHBARCLASS
-    });
+  if (!search_initialised || height != prev_height) {
+    if (!search_initialised)
+      RegisterClassA(&(WNDCLASSA){
+        .style = 0,
+        .lpfnWndProc = search_proc,
+        .cbClsExtra = 0,
+        .cbWndExtra = 0,
+        .hInstance = inst,
+        .hIcon = NULL,
+        .hCursor = NULL,
+        .hbrBackground = (HBRUSH)(COLOR_3DFACE + 1),
+        .lpszMenuName = NULL,
+        .lpszClassName = SEARCHBARCLASS
+      });
+    else {
+      DestroyWindow(search_wnd);
+    }
 
-    search_wnd = CreateWindowEx(0, SEARCHBARCLASS, "", WS_CHILD, 0, 0, 0, 0, wnd, 0, inst, NULL);
+    search_wnd = CreateWindowExA(0, SEARCHBARCLASS, "", WS_CHILD, 0, 0, 0, 0, wnd, 0, inst, NULL);
 
-    search_close_wnd = CreateWindowExW(0, L"BUTTON", L"X", WS_CHILD | WS_VISIBLE,
+    //__ label of search bar close button; not actually "localization"
+    search_close_wnd = CreateWindowExW(0, W("BUTTON"), _W("X"), WS_CHILD | WS_VISIBLE,
                                      pos_close, margin, button_width, ctrl_height,
                                      search_wnd, NULL, inst, NULL);
-    search_prev_wnd = CreateWindowExW(0, L"BUTTON", L"◀", WS_CHILD | WS_VISIBLE,
+    //__ label of search bar prev button; not actually "localization"
+    search_prev_wnd = CreateWindowExW(0, W("BUTTON"), _W("◀"), WS_CHILD | WS_VISIBLE,
                                      pos_prev, margin, button_width, ctrl_height,
                                      search_wnd, NULL, inst, NULL);
-    search_next_wnd = CreateWindowExW(0, L"BUTTON", L"▶", WS_CHILD | WS_VISIBLE,
+    //__ label of search bar next button; not actually "localization"
+    search_next_wnd = CreateWindowExW(0, W("BUTTON"), _W("▶"), WS_CHILD | WS_VISIBLE,
                                      pos_next, margin, button_width, ctrl_height,
                                      search_wnd, NULL, inst, NULL);
-    search_edit_wnd = CreateWindowEx(WS_EX_CLIENTEDGE, "EDIT", "", WS_CHILD | WS_VISIBLE | WS_TABSTOP | ES_AUTOHSCROLL,
+    search_edit_wnd = CreateWindowExA(WS_EX_CLIENTEDGE, "EDIT", "", WS_CHILD | WS_VISIBLE | WS_TABSTOP | ES_AUTOHSCROLL,
                                      0, 0, 0, 0,
                                      search_wnd, NULL, inst, NULL);
 
-    search_font = CreateFontW(ctrl_height - 4, 0, 0, 0, FW_DONTCARE, false, false, false,
+    search_font = CreateFontW(sf_height, 0, 0, 0, FW_DONTCARE, false, false, false,
                              DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
                              DEFAULT_QUALITY, FIXED_PITCH | FF_DONTCARE,
                              cfg.font.name);
@@ -232,7 +261,11 @@ win_toggle_search(bool show, bool focus)
 
     default_edit_proc = (WNDPROC)SetWindowLongPtrW(search_edit_wnd, GWLP_WNDPROC, (long)edit_proc);
 
+    if (term.results.query)
+      SetWindowTextW(search_edit_wnd, term.results.query);
+
     search_initialised = true;
+    prev_height = height;
   }
 
   if (show) {
@@ -257,25 +290,25 @@ win_toggle_search(bool show, bool focus)
 }
 
 void
-win_open_search()
+win_open_search(void)
 {
   win_toggle_search(true, true);
   win_adapt_term_size(false, false);
 }
 
-void
+static void
 win_hide_search(void)
 {
   win_toggle_search(false, false);
   win_adapt_term_size(false, false);
-  searched = false;
 }
 
-void win_update_search(void)
+void
+win_update_search(void)
 {
-    if (win_search_visible()) {
-        win_toggle_search(true, false);
-    }
+  if (win_search_visible()) {
+    win_toggle_search(true, false);
+  }
 }
 
 void
@@ -296,7 +329,8 @@ win_paint_exclude_search(HDC dc)
   ExcludeClipRect(dc, cr.left, cr.top, cr.right, cr.bottom);
 }
 
-bool win_search_visible()
+bool
+win_search_visible(void)
 {
   return IsWindowVisible(search_wnd);
 }
